from __future__ import annotations

import argparse
import cmd
import contextlib
import fnmatch
import io
import itertools
import logging
import os
import pathlib
import platform
import pydoc
import random
import re
import shlex
import shutil
import subprocess
import sys
from contextlib import contextmanager
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import TYPE_CHECKING, Any, BinaryIO, ClassVar, TextIO

from dissect.cstruct import hexdump
from flow.record import RecordOutput

from dissect.target.exceptions import (
    PluginError,
    RegistryError,
    RegistryKeyNotFoundError,
    RegistryValueNotFoundError,
    TargetError,
)
from dissect.target.helpers import cyber, fsutil, regutil
from dissect.target.helpers.logging import get_logger
from dissect.target.helpers.utils import StrEnum
from dissect.target.plugin import FunctionDescriptor, alias, arg, clone_alias
from dissect.target.target import Target
from dissect.target.tools.info import get_target_info, print_target_info
from dissect.target.tools.utils.cli import (
    catch_sigpipe,
    configure_generic_arguments,
    escape_str,
    execute_function_on_target,
    find_and_filter_plugins,
    generate_argparse_for_method,
    open_targets,
    process_generic_arguments,
)
from dissect.target.tools.utils.fs import (
    fmt_ls_colors,
    ls_scandir,
    print_ls,
    print_stat,
    print_xattr,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator


log = get_logger(__name__)
logging.lastResort = None
logging.raiseExceptions = False

try:
    import readline

    # remove `-`, `$` and `{` as an autocomplete delimeter on Linux
    # https://stackoverflow.com/questions/27288340/python-cmd-on-linux-does-not-autocomplete-special-characters-or-symbols
    readline.set_completer_delims(readline.get_completer_delims().replace("-", "").replace("$", "").replace("{", ""))

    # Fix autocomplete on macOS
    # https://stackoverflow.com/a/7116997
    if "libedit" in readline.__doc__:
        readline.parse_and_bind("bind ^I rl_complete")
    else:
        readline.parse_and_bind("tab: complete")
except ImportError:
    # Readline is not available on Windows
    log.warning("Readline module is not available")
    readline = None


def readline_escape(s: str | dict[str, str]) -> str | dict[str, str]:
    """Escape a string or values in dictionary for readline prompt.

    Used to embed terminal-specific escape sequences in prompts.

    References:
        - https://wiki.hackzine.org/development/misc/readline-color-prompt.html
        - http://stackoverflow.com/a/9468954/148845
        - RL_PROMPT_START_IGNORE = "\001"
        - RL_PROMPT_END_IGNORE = "\002"
    """
    if isinstance(s, dict):
        return {k: f"\001{v}\002" for k, v in s.items()}
    return f"\001{s}\002"


class AnsiColors(StrEnum):
    """ANSI color escape sequences."""

    # Base formatting
    RESET = "\033[0m"

    # Basic colors
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"

    # Bold colors
    BOLD_RED = "\033[1;31m"
    BOLD_GREEN = "\033[1;32m"
    BOLD_YELLOW = "\033[1;33m"
    BOLD_BLUE = "\033[1;34m"
    BOLD_MAGENTA = "\033[1;35m"
    BOLD_CYAN = "\033[1;36m"
    BOLD_WHITE = "\033[1;37m"

    @classmethod
    def as_dict(cls) -> dict[str, str]:
        """Return ANSI color escape sequences as a dictionary."""
        return {item.name: item.value for item in cls}


# ANSI color escape sequences for readline prompt
ANSI_COLORS = readline_escape(AnsiColors.as_dict()) if readline else AnsiColors.as_dict()


class ExtendedCmd(cmd.Cmd):
    """Subclassed cmd.Cmd to provide some additional features.

    Add new simple commands by implementing:
        do_<cmd>(self, line)

    Add new complex commands by implementing:
        cmd_<cmd>(self, args, stdout)

    Simple commands are plain cmd.Cmd commands. Output is generally
    presented using print().

    Complex command allow @arg decorators for argparse argument parsing.
    Additionally, complex commands allow their output to be piped to
    external commands. Complex commands should therefor always write to
    the stdout handle that is passed as argument.
    """

    CMD_PREFIX = "cmd_"
    DEFAULT_RUNCOMMANDS_FILE = None

    _runtime_aliases: ClassVar[dict[str, str]] = {}

    def __init__(self, cyber: bool = False):
        cmd.Cmd.__init__(self)
        self.debug = False
        self.cyber = cyber
        self.identchars += "."

        self.register_aliases()

    def __getattr__(self, attr: str) -> Any:
        if attr.startswith("help_"):
            _, _, command = attr.partition("_")

            def print_help(command: str, func: Callable) -> None:
                parser = generate_argparse_for_method(func, usage_tmpl=f"{command} {{usage}}")
                parser.print_help()

            try:
                func = getattr(self, self.CMD_PREFIX + command)
                return lambda: print_help(command, func)
            except AttributeError:
                pass

        return object.__getattribute__(self, attr)

    def _load_targetrc(self, path: pathlib.Path) -> None:
        """Load and execute commands from the run commands file."""
        try:
            with path.open() as fh:
                for line in fh:
                    if (line := line.strip()) and not line.startswith("#"):  # Ignore empty lines and comments
                        self.onecmd(line)
        except FileNotFoundError:
            # The .targetrc file is optional
            pass
        except Exception as e:
            log.debug("Error processing .targetrc file: %s", e)

    def _get_targetrc_path(self) -> pathlib.Path | None:
        """Get the path to the run commands file. Can return ``None`` if ``DEFAULT_RUNCOMMANDS_FILE`` is not set."""
        return pathlib.Path(self.DEFAULT_RUNCOMMANDS_FILE).expanduser() if self.DEFAULT_RUNCOMMANDS_FILE else None

    def preloop(self) -> None:
        super().preloop()
        if targetrc_path := self._get_targetrc_path():
            self._load_targetrc(targetrc_path)

    @staticmethod
    def check_compatible(target: Target) -> bool:
        return True

    def register_aliases(self) -> None:
        for name in self.get_names():
            if name.startswith(self.CMD_PREFIX):
                func = getattr(self.__class__, name)
                for alias_name in getattr(func, "__aliases__", []):
                    if not alias_name.startswith(self.CMD_PREFIX):
                        alias_name = self.CMD_PREFIX + alias_name

                    clone_alias(self.__class__, func, alias_name)

    def get_names(self) -> list[str]:
        names = cmd.Cmd.get_names(self)

        # Add fake do_ and help_ entries to get_names return value
        # This is for help output and completion
        for c in [c.partition("_")[2] for c in names if c.startswith(self.CMD_PREFIX)]:
            names.append("do_" + c)
            names.append("help_" + c)

        return names

    def _handle_command(self, line: str) -> bool | None:
        """Check whether custom handling of the cmd can be performed and if so, do it.

        If a custom handling of the cmd was performed, return the result (a boolean indicating whether the shell should
        exit). If not, return None. Can be overridden by subclasses to perform further / other 'custom command' checks.
        """
        if line == "EOF":
            return True

        # Override default command execution to first attempt complex command execution
        command, command_args_str, line = self.parseline(line)

        if hasattr(self, self.CMD_PREFIX + command):
            return self._exec_command(command, command_args_str)

        # Return None if no custom command was found to be run
        return None

    def default(self, line: str) -> bool:
        com, arg, _ = self.parseline(line)
        if com in self._runtime_aliases:
            expanded = " ".join([self._runtime_aliases[com], arg])
            return self.onecmd(expanded)

        if (should_exit := self._handle_command(line)) is not None:
            return should_exit

        # Fallback to default
        cmd.Cmd.default(self, line)
        return False

    def emptyline(self) -> None:
        """This function forces Python's cmd.Cmd module to behave like a regular shell.

        When entering an empty command, the cmd module will by default repeat the previous command.
        By defining an empty ``emptyline`` function we make sure no command is executed instead.

        References:
            - https://stackoverflow.com/a/16479030
            - https://github.com/python/cpython/blob/3.12/Lib/cmd.py#L10
        """

    def _exec(self, func: Callable[[list[str], TextIO], bool], command_args_str: str, no_cyber: bool = False) -> bool:
        """Command execution helper that chains initial command, piped subprocesses, and output redirection together."""

        argparts = []
        if command_args_str is not None:
            argparts = arg_str_to_arg_list(command_args_str)

        # Enforce that output redirection (>) only appears after the last pipe (|)
        redirect_idx = None
        redirect_file = None
        if ">" in argparts:
            redirect_indices = [i for i, v in enumerate(argparts) if v == ">"]
            # Only support single output redirection
            if len(redirect_indices) > 1:
                print("Syntax error: multiple output redirections specified")
                return False
            redirect_idx = redirect_indices[0]
            if redirect_idx + 1 >= len(argparts):
                print("Syntax error: missing filename after '>'")
                return False
            # If there are pipes, > must be after the last |
            if "|" in argparts[redirect_idx:]:
                print("Syntax error: output redirection must come after the last pipe")
                return False
            redirect_file = argparts[redirect_idx + 1]
            # Remove redirect from argparts
            argparts = argparts[:redirect_idx]

        # Handle pipes
        if "|" in argparts:
            pipeidx = argparts.index("|")
            argparts, pipeparts = argparts[:pipeidx], argparts[pipeidx + 1 :]
            try:
                # If redirect, open file for writing and pass as stdout
                if redirect_file:
                    with (
                        Path(redirect_file).open("wb") as f,
                        io.TextIOWrapper(f, encoding="utf-8") as tf,
                        build_pipe(pipeparts, pipe_stdout=tf) as (pipe_stdin, _),
                    ):
                        return func(argparts, pipe_stdin)
                else:
                    with build_pipe_stdout(pipeparts) as pipe_stdin:
                        return func(argparts, pipe_stdin)
            except OSError as e:
                # in case of a failure in a subprocess
                print(e)
                return False
        else:
            ctx = contextlib.nullcontext()
            if self.cyber and not no_cyber:
                ctx = cyber.cyber(color=None, run_at_end=True)

            # If redirect without pipes, open file for writing and pass as stdout
            if redirect_file:
                with (
                    Path(redirect_file).open("wb") as f,
                    io.TextIOWrapper(f, encoding="utf-8") as tf,
                    ctx,
                ):
                    return func(argparts, tf)
            else:
                with ctx:
                    return func(argparts, sys.stdout)

    def _exec_command(self, command: str, command_args_str: str) -> bool:
        """Command execution helper for ``cmd_`` commands."""
        cmdfunc = getattr(self, self.CMD_PREFIX + command)
        argparser = generate_argparse_for_method(cmdfunc, usage_tmpl=f"{command} {{usage}}")

        def _exec_(argparts: list[str], stdout: TextIO) -> bool:
            try:
                args = argparser.parse_args(argparts)
            except SystemExit:
                return False
            return cmdfunc(args, stdout)

        # These commands enter a subshell, which doesn't work well with cyber
        no_cyber = cmdfunc.__func__ in (TargetCli.cmd_registry, TargetCli.cmd_enter)
        return self._exec(_exec_, command_args_str, no_cyber)

    def do_man(self, line: str) -> bool:
        """alias for help"""
        self.do_help(line)
        return False

    def complete_man(self, *args: list[str]) -> list[str]:
        return cmd.Cmd.complete_help(self, *args)

    def do_unalias(self, line: str) -> bool:
        """delete runtime alias"""
        aliases = list(shlex.shlex(line, posix=True))
        for aliased in aliases:
            if aliased in self._runtime_aliases:
                del self._runtime_aliases[aliased]
            else:
                print(f"alias {aliased} not found")
        return False

    def do_alias(self, line: str) -> bool:
        """create a runtime alias"""
        args = list(shlex.shlex(line, posix=True))

        if not args:
            for aliased, command in self._runtime_aliases.items():
                print(f"alias {aliased}={command}")
            return False

        while args:
            alias_name = args.pop(0)
            try:
                equals = args.pop(0)
                # our parser works different, so we have to stop this
                if equals != "=":
                    raise RuntimeError("Token not allowed")
                expanded = args.pop(0) if args else ""  # this is how it works in bash
                self._runtime_aliases[alias_name] = expanded
            except IndexError:
                if alias_name in self._runtime_aliases:
                    print(f"alias {alias_name}={self._runtime_aliases[alias_name]}")
                else:
                    print(f"alias {alias_name} not found")

        return False

    def do_clear(self, line: str) -> bool:
        """clear the terminal screen"""
        os.system("cls||clear")
        return False

    def do_cls(self, line: str) -> bool:
        """alias for clear"""
        return self.do_clear(line)

    def do_exit(self, line: str) -> bool:
        """exit shell"""
        return True

    def do_cyber(self, line: str) -> bool:
        """cyber"""
        self.cyber = not self.cyber
        word, color = {False: ("D I S E N", cyber.Color.RED), True: ("E N", cyber.Color.YELLOW)}[self.cyber]
        with cyber.cyber(color=color):
            print(f"C Y B E R - M O D E - {word} G A G E D")
        return False

    def do_debug(self, line: str) -> bool:
        """toggle debug mode"""
        self.debug = not self.debug
        if self.debug:
            print("Debug mode on")
        else:
            print("Debug mode off")
        return False


class TargetCmd(ExtendedCmd):
    DEFAULT_HISTFILE = "~/.dissect_history"
    DEFAULT_HISTFILESIZE = 10_000
    DEFAULT_HISTDIR = None
    DEFAULT_HISTDIRFMT = ".dissect_history_{uid}_{target}"
    DEFAULT_RUNCOMMANDS_FILE = "~/.targetrc"
    CONFIG_KEY_RUNCOMMANDS_FILE = "TARGETRCFILE"

    def __init__(self, target: Target):
        self.target = target

        # history file
        self.histfilesize = getattr(target._config, "HISTFILESIZE", self.DEFAULT_HISTFILESIZE)
        self.histdir = getattr(target._config, "HISTDIR", self.DEFAULT_HISTDIR)

        if self.histdir:
            self.histdirfmt = getattr(target._config, "HISTDIRFMT", self.DEFAULT_HISTDIRFMT)
            self.histfile = pathlib.Path(self.histdir).resolve() / pathlib.Path(
                self.histdirfmt.format(uid=os.getuid(), target=target.name)
            )
        else:
            self.histfile = pathlib.Path(getattr(target._config, "HISTFILE", self.DEFAULT_HISTFILE)).expanduser()

        # prompt format
        self.prompt_ps1 = "{BOLD_GREEN}{base}{RESET}:{BOLD_BLUE}{cwd}{RESET}$ "
        if ps1 := getattr(target._config, "PS1", None):
            if "{cwd}" in ps1 and "{base}" in ps1:
                self.prompt_ps1 = ps1
            else:
                self.target.log.warning("{cwd} and {base} were not set inside PS1, using the default prompt")

        elif getattr(target._config, "NO_COLOR", None) or os.getenv("NO_COLOR"):
            self.prompt_ps1 = "{base}:{cwd}$ "

        super().__init__(self.target.props.get("cyber"))

    def _get_targetrc_path(self) -> pathlib.Path:
        """Get the path to the run commands file."""

        return pathlib.Path(
            getattr(self.target._config, self.CONFIG_KEY_RUNCOMMANDS_FILE, self.DEFAULT_RUNCOMMANDS_FILE)
        ).expanduser()

    def preloop(self) -> None:
        super().preloop()
        if readline and self.histfile.exists():
            try:
                readline.read_history_file(self.histfile)
            except Exception as e:
                log.debug("Error reading history file: %s", e)

    def postloop(self) -> None:
        if readline:
            readline.set_history_length(self.histfilesize)
            try:
                readline.write_history_file(self.histfile)
            except Exception as e:
                log.debug("Error writing history file: %s", e)

    def _handle_command(self, line: str) -> bool | None:
        if (should_exit := super()._handle_command(line)) is not None:
            return should_exit

        # The parent class has already attempted complex command execution, we now attempt target plugin command
        # execution
        command, command_args_str, line = self.parseline(line)

        if functions := list(find_and_filter_plugins(command, self.target)):
            return self._exec_target(functions, command_args_str)

        # We didn't execute a function on the target
        return None

    def _exec_target(self, funcs: list[FunctionDescriptor], command_args_str: str) -> bool:
        """Command exection helper for target plugins."""

        def _exec_(argparts: list[str], stdout: TextIO) -> None:
            try:
                output, value = execute_function_on_target(self.target, func, argparts)
            except SystemExit:
                return

            if output == "record":
                # if the command results are piped to another process,
                # the process will receive Record objects
                if stdout is sys.stdout:
                    for entry in value:
                        print(entry, file=stdout)
                else:
                    stdout = stdout.buffer

                    rs = RecordOutput(stdout)
                    for entry in value:
                        rs.write(entry)
            elif output == "yield":
                for entry in value:
                    print(entry, file=stdout)
            elif output == "none":
                return
            else:
                print(value, file=stdout)

        # What in the variable hoisting is this
        for func in funcs:  # noqa: B007
            try:
                self._exec(_exec_, command_args_str)
            except PluginError:  # noqa: PERF203
                if self.debug:
                    raise
                self.target.log.exception("Plugin error")

        # Keep the shell open
        return False

    def do_python(self, line: str) -> bool:
        """drop into a Python shell"""
        python_shell([self.target])
        return False


class TargetHubCli(cmd.Cmd):
    """Hub Cli for interacting with multiple targets."""

    prompt = "dissect> "
    doc_header = (
        "Target Hub\n"
        "==========\n"
        "List and enter targets by using 'list' and 'enter'.\n\n"
        "Documented commands (type help <topic>):"
    )

    def __init__(self, targets: list[Target], cli: TargetCmd):
        cmd.Cmd.__init__(self)
        self.targets = targets
        self._targetcli = cli
        self._names = [t.name for t in targets]
        self._names_lower = [h.lower() for h in self._names]

        self._clicache = {}

    def default(self, line: str) -> bool:
        if line == "EOF":
            return True

        cmd.Cmd.default(self, line)
        return False

    def emptyline(self) -> None:
        pass

    def do_exit(self, line: str) -> bool:
        """exit shell"""
        return True

    def do_list(self, line: str) -> bool:
        """list the loaded targets"""
        print("\n".join([f"{i:2d}: {e}" for i, e in enumerate(self._names)]))
        return False

    def do_enter(self, line: str) -> bool:
        """enter a target by number or name"""

        if line.isdigit():
            idx = int(line)
        else:
            try:
                idx = self._names_lower.index(line.lower())
            except ValueError:
                print("Unknown name")
                return False

        if idx >= len(self.targets):
            print("Unknown target")
            return False

        try:
            cli = self._clicache[idx]
        except KeyError:
            target = self.targets[idx]
            if not self._targetcli.check_compatible(target):
                return False

            cli = self._targetcli(self.targets[idx])
            self._clicache[idx] = cli

        print(f"Entering {idx}: {self._names[idx]}")
        run_cli(cli)
        return False

    def complete_enter(self, text: str, line: str, begidx: int, endidx: int) -> list[str]:
        if not text:
            return self._names[:]

        compl = [h for h in self._names if h.lower().startswith(text.lower())]

        # Also provide completion for multidigit numbers
        if text.isdigit():
            compl.extend([str(i) for i in range(len(self.targets)) if str(i).startswith(text)])

        return compl

    def do_python(self, line: str) -> bool:
        """drop into a Python shell"""
        python_shell(self.targets)
        return False


class TargetCli(TargetCmd):
    """CLI for interacting with a target and browsing the filesystem."""

    def __init__(self, target: Target):
        self.prompt_base = _target_name(target)

        TargetCmd.__init__(self, target)
        self._clicache = {}
        # Force to root, using `chdir` causes `None` to propagate throughout the class methods.
        self.cwd = self.target.fs.path("/")

    @property
    def prompt(self) -> str:
        return self.prompt_ps1.format(base=self.prompt_base, cwd=self.cwd, **ANSI_COLORS)

    def completedefault(self, text: str, line: str, begidx: int, endidx: int) -> list[str]:
        path = self.resolve_path(line[:begidx].rsplit(" ")[-1])
        textlower = text.lower()

        suggestions = []
        for fpath, fname in ls_scandir(path):
            if not fname.lower().startswith(textlower):
                continue

            # Add a trailing slash to directories, to allow for easier traversal of the filesystem
            suggestion = f"{fname}/" if fpath.is_dir() else fname
            suggestions.append(suggestion)
        return suggestions

    def resolve_path(self, path: str) -> fsutil.TargetPath:
        if not path:
            return self.cwd

        if isinstance(path, fsutil.TargetPath):
            return path

        path = fsutil.abspath(path, cwd=str(self.cwd), alt_separator=self.target.fs.alt_separator)
        return self.target.fs.path(path)

    def resolve_glob_path(self, path: str) -> Iterator[fsutil.TargetPath]:
        path = self.resolve_path(path)
        if path.exists():
            yield path
        else:
            # Strip the leading '/' as non-relative patterns are unsupported as glob patterns.
            glob_path = str(path).lstrip("/")
            try:
                for path in self.target.fs.path("/").glob(glob_path):
                    yield path
            except ValueError as e:
                # The generator returned by glob() will raise a
                # ValueError if the '**' glob is not used as an entire path
                # component
                print(e)

    def check_file(self, path: str) -> fsutil.TargetPath | None:
        path = self.resolve_path(path)

        if not path.exists():
            print(f"{path}: No such file")
            return None

        # Check a special case where a path can be both a file and directory (e.g. NTDS.dit)
        # We need to check this on the entry, as the path methods can't detect this because of how stat.S_IS* works
        entry = path.get()
        if entry.is_file() and entry.is_dir():
            return path

        if path.is_dir():
            print(f"{path}: Is a directory")
            return None

        if not path.is_file():
            print(f"{path}: Not a file")
            return None

        return path

    def check_dir(self, path: str) -> fsutil.TargetPath | None:
        path = self.resolve_path(path)
        if not path.exists():
            print(f"{path}: No such directory")
            return None

        if path.is_file():
            print(f"{path}: Is a file")
            return None

        if not path.is_dir():
            print(f"{path}: Not a directory")
            return None

        return path

    def check_path(self, path: str) -> fsutil.TargetPath | None:
        path = self.resolve_path(path)
        if not path.exists():
            print(f"{path}: No such file or directory")
            return None

        return path

    def chdir(self, path: str) -> None:
        """Change directory to the given path."""
        if dir := self.check_dir(path):
            self.cwd = dir

    def do_cd(self, line: str) -> bool:
        """change directory"""
        self.chdir(line)
        return False

    def do_pwd(self, line: str) -> bool:
        """print current directory"""
        print(self.cwd)
        return False

    def do_disks(self, line: str) -> bool:
        """print target disks"""
        for d in self.target.disks:
            print(str(d))
        return False

    def do_volumes(self, line: str) -> bool:
        """print target volumes"""
        for v in self.target.volumes:
            print(str(v))
        return False

    def do_filesystems(self, line: str) -> bool:
        """print target filesystems"""
        for fs in self.target.filesystems:
            print(str(fs))
        return False

    def do_mounts(self, line: str) -> bool:
        """print target mounts"""
        for mount, fs in self.target.fs.mounts.items():
            print(f"<Mount fs={fs.__type__!r} path={mount!r}>")
        return False

    def do_info(self, line: str) -> bool:
        """print target information"""
        target_info = get_target_info(self.target)
        print_target_info(self.target, target_info)
        return False

    def do_reload(self, line: str) -> bool:
        """reload the target"""
        self.target = self.target.reload()
        if self.cwd:
            self.chdir(str(self.cwd))  # self.cwd has reference into the old target :/
        return False

    @arg("path", nargs="?")
    @arg("-l", action="store_true")
    @arg("-a", "--all", action="store_true")  # ignored but included for proper argument parsing
    @arg("-h", "--human-readable", action="store_true")
    @arg("-R", "--recursive", action="store_true", help="recursively list subdirectories encountered")
    @arg("-c", action="store_true", dest="use_ctime", help="show time when file status was last changed")
    @arg("-u", action="store_true", dest="use_atime", help="show time of last access")
    @alias("l")
    @alias("dir")
    def cmd_ls(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """list directory contents"""

        path = self.resolve_path(args.path)

        if args.use_ctime and args.use_atime:
            print("can't specify -c and -u at the same time")
            return False

        if not path:
            return False

        if not path.exists():
            print(f"ls: cannot access {path}: No such file or directory")
            return False

        # Disable color if output is redirected to a file
        use_color = False
        if hasattr(stdout, "isatty") and stdout.isatty():
            use_color = True
        if os.getenv("NO_COLOR") in ["1", "True", "true"]:
            use_color = False

        print_ls(
            path,
            0,
            stdout,
            args.l,
            args.human_readable,
            args.recursive,
            args.use_ctime,
            args.use_atime,
            color=use_color,
        )
        return False

    @arg("path", nargs="?")
    def cmd_ll(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """alias for ls -la"""
        args = extend_args(args, self.cmd_ls)
        args.l = True
        args.a = True
        return self.cmd_ls(args, stdout)

    @arg("path", nargs="?")
    def cmd_tree(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """alias for ls -R"""
        args = extend_args(args, self.cmd_ls)
        args.recursive = True
        return self.cmd_ls(args, stdout)

    @arg("path", nargs="?")
    @arg("-name", default="*", help="path to match with")
    @arg("-iname", help="like -name, but the match is case insensitive")
    @arg("-atime", type=int, help="file was last accessed n*24 hours ago")
    @arg("-mtime", type=int, help="file was last modified n*24 hours ago")
    @arg("-ctime", type=int, help="file (windows) or metadata (unix) was last changed n*24 hours ago")
    @arg("-btime", type=int, help="file was born n*24 hours ago (ext4)")
    def cmd_find(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """search for files in a directory hierarchy"""
        path = self.resolve_path(args.path)
        if not path or not self.check_dir(path):
            return False

        matches = []
        now = datetime.now(tz=timezone.utc)
        do_time_compare = any([args.mtime, args.atime])

        if args.iname:
            pattern = re.compile(fnmatch.translate(args.iname), re.IGNORECASE)
            for f in path.rglob("*"):
                if pattern.match(f.name):
                    matches.append(f) if do_time_compare else print(f, file=stdout)
        else:
            for f in path.rglob(args.name):
                matches.append(f) if do_time_compare else print(f, file=stdout)

        def compare(now: datetime, then_ts: float, offset: int) -> bool:
            then = datetime.fromtimestamp(then_ts, tz=timezone.utc)
            return now - timedelta(hours=offset * 24) > then

        if do_time_compare:
            for f in matches:
                s = f.lstat()

                if args.mtime and compare(now, s.st_mtime, offset=args.mtime):
                    continue

                if args.atime and compare(now, s.st_atime, offset=args.atime):
                    continue

                if args.ctime and compare(now, s.st_ctime, offset=args.ctime):
                    continue

                if args.btime and compare(now, s.st_birthtime, offset=args.btime):
                    continue

                print(f, file=stdout)

        return False

    @arg("path")
    @arg("-L", "--dereference", action="store_true")
    def cmd_stat(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """display file status"""
        path = self.resolve_path(args.path)
        if not path or not self.check_path(path):
            return False

        print_stat(path, stdout, args.dereference)
        return False

    @arg("path")
    @arg("-d", "--dump", action="store_true")
    @arg("-R", "--recursive", action="store_true")
    @alias("getfattr")
    def cmd_attr(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """display file attributes"""
        path = self.resolve_path(args.path)
        if not path or not self.check_path(path):
            return False

        try:
            if attr := path.get().attr():
                print_xattr(path.name, attr, stdout)
        except Exception:
            pass

        if args.recursive:
            for child in path.rglob("*"):
                try:
                    if child_attr := child.get().attr():
                        print_xattr(child, child_attr, stdout)
                        print()
                except Exception:  # noqa: PERF203
                    pass

        return False

    @arg("path")
    def cmd_file(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """determine file type"""
        if not (path := self.check_file(args.path)):
            return False

        p = subprocess.Popen(["file", "-"], stdin=subprocess.PIPE, stdout=subprocess.PIPE)

        with path.open() as fh:
            # We could just alias this to cat <path> | file -, but this is slow for large files
            # This way we can explicitly limit to just 512 bytes
            p.stdin.write(fh.read(512))

        p.stdin.close()
        p.wait()

        filetype = p.stdout.read().decode().split(":", 1)[1].strip()
        print(f"{path}: {filetype}", file=stdout)

        return False

    @arg("path", nargs="+")
    @arg("-o", "--out", default=".")
    @arg("-v", "--verbose", action="store_true")
    def cmd_save(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """save a common file or directory to the host filesystem"""
        dst_path = pathlib.Path(args.out).resolve()

        if len(args.path) > 1 and not dst_path.is_dir():
            # Saving multiple items to a non-directory is generally not very
            # useful. This also prevents multiple error messages if some of the
            # items are directories.
            print(f"{dst_path}: cannot save multiple files, destination is not a directory")

        def log_saved_path(src_path: pathlib.Path, dst_path: pathlib.Path) -> None:
            if args.verbose:
                print(f"{src_path} -> {dst_path}")

        def get_diverging_path(path: pathlib.Path, reference_path: pathlib.Path) -> pathlib.Path:
            """Get the part of path where it diverges from reference_path."""
            diverging_path = pathlib.Path()

            for diff_idx, path_part in enumerate(reference_path.parts):
                if path_part != path.parts[diff_idx]:
                    diverging_parts = path.parts[diff_idx:]
                    if diverging_parts:
                        for part in diverging_parts:
                            diverging_path = diverging_path.joinpath(pathlib.Path(part))
                    break

            return diverging_path

        def save_path(
            src_path: pathlib.Path, dst_path: pathlib.Path, create_dst_subdir: pathlib.Path | None = None
        ) -> None:
            """Save a common file or directory in src_path to dst_path.

            If src_path is a file, dst_path can be either a directory or a file
            (the latter case will rename the source file).

            If src_path is a directory, dst_path is only allowed to be a
            directory.

            Symlinks in src_path are not explicitly resolved, if they point to
            a file or directory they will be saved under the symlinked name.

            If create_dst_subdir is specified, this sub directory is first
            created below dst_path (if dst_path is a directory) after which the
            src_path is saved there. This can be useful to recreate a part of
            the directory hierarchy in src_path.
            """

            src_name = src_path.name
            src_path = src_path.resolve()
            if src_path.is_dir():
                # Directories can only be copied if the destination is a
                # directory.
                if not dst_path.exists():
                    print(f"{dst_path}: destination directory does not exist")
                    return
                if not dst_path.is_dir():
                    print(f"{src_path}: cannot save directory into non-directory {dst_path}")
                    return
                if create_dst_subdir:
                    dst_path = dst_path.joinpath(create_dst_subdir)
                dst_path = dst_path.joinpath(src_name)
                dst_path.mkdir(parents=True, exist_ok=True)
                log_saved_path(src_path, dst_path)
                for path in src_path.glob("*"):
                    save_path(path, dst_path)

            elif src_path.is_file():
                # Files will overwrite existing destination files, if they
                # exist, or be created if they don't.
                if not dst_path.exists() and not dst_path.parent.is_dir():
                    print(f"{dst_path.parent}: destination directory does not exist")
                    return
                if dst_path.exists():
                    if dst_path.is_dir():
                        if create_dst_subdir:
                            dst_path = dst_path.joinpath(create_dst_subdir)
                            dst_path.mkdir(parents=True, exist_ok=True)
                        dst_path = dst_path.joinpath(src_name)
                    elif not dst_path.is_file():
                        # Saving over special files like device or pipe files
                        # won't work (and is not useful).
                        print(f"{src_path}: cannot save file as non-file {dst_path}")
                        return

                with dst_path.open(mode="wb") as dst_fh:
                    shutil.copyfileobj(src_path.open(), dst_fh)
                log_saved_path(src_path, dst_path)

            else:
                if src_path.exists():
                    print(f"{src_path}: not a file or directory")
                else:
                    print(f"{src_path}: no such file or directory")

        for path in args.path:
            # This normalizes /'s and \'s, resolves . and ..  and makes the
            # path absolute, but does not resolve symlinks like pathlib's
            # Path.resolve(). This is needed as we need the original filename
            # for saving.
            src_paths = self.resolve_glob_path(path)
            reference_path = self.resolve_path(path)
            # See if the path has globs in it which resolve to multiple files.
            # If so the dst_path needs to be a directory to properly save all
            # the files resulting from globbing.
            try:
                first_src_path = next(src_paths)
            except StopIteration:
                print(f"{path}: No such file or directory")
                return False

            try:
                second_src_path = next(src_paths)
            except StopIteration:
                # We use the parent of the diverging path as the final part
                # (if there is any) is the file or directory we want to
                # save.
                extra_dir = get_diverging_path(first_src_path, reference_path).parent
                save_path(first_src_path, dst_path, create_dst_subdir=extra_dir)
            else:
                if not dst_path.is_dir():
                    # Saving multiple items to a non-directory is generally not very
                    # useful. This also prevents multiple error messages if some of the
                    # items are directories.
                    print(f"{dst_path}: cannot save multiple files, destination is not a directory")
                else:
                    for src_path in itertools.chain([first_src_path, second_src_path], src_paths):
                        extra_dir = get_diverging_path(src_path, reference_path).parent
                        save_path(src_path, dst_path, create_dst_subdir=extra_dir)

        return False

    @arg("path")
    @alias("type")
    def cmd_cat(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """print file content"""
        paths = list(self.resolve_glob_path(args.path))

        if not paths:
            print(f"{args.path}: No such file or directory")
            return False

        stdout = stdout.buffer
        for path in paths:
            if not (path := self.check_file(path)):
                continue

            with path.open() as fh:
                shutil.copyfileobj(fh, stdout)
            stdout.flush()
        print()
        return False

    @arg("path")
    def cmd_zcat(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """print file content from compressed files"""
        paths = list(self.resolve_glob_path(args.path))

        if not paths:
            print(f"{args.path}: No such file or directory")
            return False

        stdout = stdout.buffer
        for path in paths:
            if not (path := self.check_file(path)):
                continue

            with fsutil.open_decompress(path) as fh:
                shutil.copyfileobj(fh, stdout)
            stdout.flush()

        return False

    @arg("path")
    @arg("-n", "--length", type=int, default=16 * 20, help="amount of bytes to read")
    @arg("-s", "--skip", type=int, default=0, help="skip offset bytes from the beginning")
    @arg("-p", "--hex", action="store_true", help="output in plain hexdump style")
    @arg("-C", "--canonical", action="store_true")
    @alias("xxd")
    def cmd_hexdump(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """print a hexdump of file(s)"""
        paths = list(self.resolve_glob_path(args.path))
        if not paths:
            print(f"{args.path}: No such file or directory")
            return False

        for path in paths:
            if len(paths) > 1:
                print(f"[{path}]", file=stdout)
            with path.open("rb") as fh:
                if args.skip > 0:
                    fh.seek(args.skip + 1)

                if args.hex:
                    print(fh.read(args.length).hex(), file=stdout)
                else:
                    print(hexdump(fh.read(args.length), output="string"), file=stdout)

        return False

    @arg("path")
    @alias("digest")
    @alias("shasum")
    def cmd_hash(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """print the MD5, SHA1 and SHA256 hashes of a file"""
        if not (path := self.check_file(args.path)):
            return False

        md5, sha1, sha256 = path.get().hash()
        print(f"MD5:\t{md5}\nSHA1:\t{sha1}\nSHA256:\t{sha256}", file=stdout)

        return False

    @arg("path")
    def cmd_md5sum(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """print the MD5 checksum of a file provided by a path"""
        if not (path := self.check_file(args.path)):
            return False

        (md5,) = path.get().hash(["md5"])
        print(f"{md5}  {path!s}", file=stdout)
        return False

    @arg("path")
    def cmd_sha1sum(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """print the SHA1 checksum of a file provided by a path"""
        if not (path := self.check_file(args.path)):
            return False

        (sha1,) = path.get().hash(["sha1"])
        print(f"{sha1}  {path!s}", file=stdout)
        return False

    @arg("path")
    def cmd_sha256sum(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """print the SHA256 checksum of a file provided by a path"""
        if not (path := self.check_file(args.path)):
            return False

        (sha256,) = path.get().hash(["sha256"])
        print(f"{sha256}  {path!s}", file=stdout)
        return False

    @arg("path")
    @alias("head")
    @alias("more")
    def cmd_less(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """open the first 10 MB of a file with less"""
        if not (path := self.check_file(args.path)):
            return False

        with path.open("rt", errors="ignore") as fh:
            pydoc.pager(fh.read(10 * 1024 * 1024))

        return False

    @arg("path")
    @alias("zhead")
    @alias("zmore")
    def cmd_zless(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """open the first 10 MB of a compressed file with zless"""
        if not (path := self.check_file(args.path)):
            return False

        with fsutil.open_decompress(path, "rt") as fh:
            pydoc.pager(fh.read(10 * 1024 * 1024))

        return False

    @arg("path", nargs="+")
    def cmd_readlink(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """print resolved symbolic links or canonical file names"""
        for path in args.path:
            path = self.resolve_path(path)
            if not path.is_symlink():
                continue

            print(path.get().readlink(), file=stdout)

        return False

    @arg("path", nargs="?", help="load a hive from the given path")
    def cmd_registry(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """drop into a registry shell"""
        hive = None

        clikey = "registry"
        if args.path:
            if not (path := self.check_file(args.path)):
                return False

            hive = regutil.RegfHive(path)
            clikey = f"registry_{path}"

        try:
            cli = self._clicache[clikey]
        except KeyError:
            if not hive and not RegistryCli.check_compatible(self.target):
                return False

            cli = RegistryCli(self.target, hive)
            self._clicache[clikey] = cli

        run_cli(cli)
        print()
        return False

    @arg("targets", metavar="TARGETS", nargs="*", help="targets to load")
    @arg("-p", "--python", action="store_true", help="(I)Python shell")
    @arg("-r", "--registry", action="store_true", help="registry shell")
    def cmd_enter(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        """load one or more files as sub-targets and drop into a sub-shell"""
        paths = [self.resolve_path(path) for path in args.targets]

        if args.python:
            # Quick path that doesn't require CLI caching
            open_shell(list(open_targets(args)), args.python, args.registry)
            return False

        clikey = tuple(str(path) for path in paths)
        try:
            cli = self._clicache[clikey]
        except KeyError:
            targets = list(Target.open_all(paths))

            cli = create_cli(targets, RegistryCli if args.registry else TargetCli)
            if not cli:
                return False

            self._clicache[clikey] = cli

        run_cli(cli)
        print()
        return False


class RegistryCli(TargetCmd):
    """CLI for browsing the registry."""

    # Registry shell is incompatible with default shell, so override the default rc file and config key
    DEFAULT_RUNCOMMANDS_FILE = "~/.targetrc.registry"
    CONFIG_KEY_RUNCOMMANDS_FILE = "TARGETRCFILE_REGISTRY"

    def __init__(self, target: Target, registry: regutil.RegfHive | None = None):
        self.prompt_base = _target_name(target)

        TargetCmd.__init__(self, target)

        self.registry = registry or target.registry
        self.cwd = None
        self.chdir("\\")

    @staticmethod
    def check_compatible(target: Target) -> bool:
        if not target.has_function("registry"):
            print("ERROR: Target doesn't have a registry")
            return False
        return True

    @property
    def prompt(self) -> str:
        return "(registry) " + self.prompt_ps1.format(base=self.prompt_base, cwd=self.cwd, **ANSI_COLORS)

    def completedefault(self, text: str, line: str, begidx: int, endidx: int) -> list[str]:
        path = line[:begidx].rsplit(" ")[-1]
        return [fname for _, fname in self.scandir(path) if fname.lower().startswith(text.lower())]

    def resolve_key(self, path: str) -> regutil.RegistryKey:
        if isinstance(path, regutil.RegistryKey):
            return path

        path = f"{self.cwd}\\{path}" if path and not path.startswith("\\") else path or self.cwd
        path = path.replace("\\\\", "\\")
        return self.registry.key(path.strip("\\"))

    def check_key(self, path: str) -> regutil.RegistryKey:
        try:
            return self.resolve_key(path)
        except RegistryError:
            print(f"{path}: No such subkey")

    def check_value(self, path: str) -> regutil.RegistryValue:
        path, _, value = path.rpartition("\\")
        try:
            key = self.resolve_key(path)
            return key.value(value)
        except RegistryKeyNotFoundError:
            print(f"{path}: No such subkey")
        except RegistryValueNotFoundError:
            print(f"{path}\\{value}: No such value")

    def chdir(self, path: str) -> None:
        if not path.startswith("\\"):
            path = f"{self.cwd}\\{path}"

        if self.check_key(path):
            self.cwd = "\\" + path.strip("\\")

    def scandir(self, path: str, color: bool = False) -> list[tuple[regutil.RegistryKey | regutil.RegistryValue, str]]:
        try:
            key = self.resolve_key(path)
        except RegistryError:
            return []

        r = [
            *((s, fmt_ls_colors("di", s.name) if color else s.name) for s in key.subkeys()),
            *((v, fmt_ls_colors("fi", v.name) if color else v.name) for v in key.values()),
        ]

        r.sort(key=lambda e: e[0].name)
        return r

    def do_cd(self, line: str) -> bool:
        """change subkey"""
        if line == "..":
            try:
                self.resolve_key(self.cwd + "\\..")
            except RegistryError:
                self.do_up(line)
                return False

        self.chdir(line)
        return False

    def do_up(self, line: str) -> bool:
        """go up a subkey"""
        parent = self.cwd.rpartition("\\")[0]
        if not parent:
            parent = "\\"
        self.chdir(parent)
        return False

    def do_pwd(self, line: str) -> bool:
        """print current path"""
        print(self.cwd.lstrip("\\"))
        return False

    def do_recommend(self, line: str) -> bool:
        """recommend a key"""
        print(random.choice([name for _, name in self.scandir(None)]))
        return False

    @arg("path", nargs="?")
    def cmd_ls(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        key = self.check_key(args.path)
        if not key:
            return False

        r = self.scandir(key, color=True)
        print("\n".join([name for _, name in r]), file=stdout)
        return False

    @arg("value")
    def cmd_cat(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        value = self.check_value(args.value)
        if not value:
            return False

        print(repr(value.value), file=stdout)
        return False

    @arg("value")
    @arg("-p", "--hex", action="store_true")
    @alias("xxd")
    def cmd_hexdump(self, args: argparse.Namespace, stdout: TextIO) -> bool:
        value = self.check_value(args.value)
        if not value:
            return False

        if args.hex:
            print(value.value.hex(), file=stdout)
        else:
            print(hexdump(value.value, output="string"), file=stdout)

        return False


def arg_str_to_arg_list(args: str) -> list[str]:
    """Convert a commandline string to a list of command line arguments."""
    lexer = shlex.shlex(args, posix=True, punctuation_chars=True)
    lexer.wordchars += "$"
    lexer.whitespace_split = True
    return list(lexer)


def extend_args(args: argparse.Namespace, func: Callable) -> argparse.Namespace:
    """Extend the arguments of the given ``func`` with the provided ``argparse.Namespace``."""
    for short, kwargs in func.__args__:
        name = kwargs.get("dest", short[-1]).lstrip("-").replace("-", "_")
        if not hasattr(args, name):
            setattr(args, name, None)
    return args


def _target_name(target: Target) -> str:
    """Return a printable FQDN target name for cmd.Cmd base prompts."""

    if target.has_function("domain") and target.domain:
        return escape_str(f"{target.name}.{target.domain}")

    return escape_str(target.name)


@contextmanager
def build_pipe(pipe_parts: list[str], pipe_stdout: int = subprocess.PIPE) -> Iterator[tuple[TextIO, BinaryIO]]:
    """Generator context manager that chains piped subprocessess and
    returns a tuple (chain input stream, chain output stream).

    On context exit the generator will close the input stream and wait for
    the subprocessess to finish.
    """
    if not pipe_parts:
        raise ValueError("No pipe components provided")

    proc_args = []
    current_stdin = subprocess.PIPE
    pipe_stdin = None

    for part in pipe_parts:
        if part == "|":
            proc = subprocess.Popen(proc_args, stdin=current_stdin, stdout=subprocess.PIPE)
            if not pipe_stdin:
                pipe_stdin = proc.stdin

            proc_args = []
            current_stdin = proc.stdout
        else:
            proc_args.append(part)

    if not proc_args:
        # fail gracefully by piping stdin into stdout if command ends
        # with hanging `|`
        proc_args = ["cat"]

    # last process in a pipe
    proc = subprocess.Popen(proc_args, stdin=current_stdin, stdout=pipe_stdout)
    if not pipe_stdin:
        pipe_stdin = proc.stdin

    pipe_stdout = proc.stdout

    # turn a byte stream into a text stream
    pipe_txt_stdin = io.TextIOWrapper(pipe_stdin)
    try:
        yield pipe_txt_stdin, pipe_stdout
    finally:
        try:
            pipe_txt_stdin.close()
        except IOError:
            # Ignore errors if pipe closes prematurely
            pass
    proc.wait()


@contextmanager
def build_pipe_stdout(pipe_parts: list[str]) -> Iterator[TextIO]:
    """Generator context manager that chains piped subprocessess, with a chain's
    outgoing stream configured to be parent's stdout.

    Generator returns a chain's input stream from `build_pipe` generator.
    """
    with build_pipe(pipe_parts, pipe_stdout=None) as (pipe_stdin, _):
        yield pipe_stdin


def open_shell(
    targets: list[Target], python: bool = False, registry: bool = False, commands: list[str] | None = None
) -> None:
    """Helper method for starting a regular, Python or registry shell for one or multiple targets."""
    if python:
        python_shell(targets, commands=commands)
    else:
        cli_cls = RegistryCli if registry else TargetCli
        target_shell(targets, cli_cls=cli_cls, commands=commands)


def target_shell(targets: list[Target], cli_cls: type[TargetCmd], commands: list[str] | None = None) -> None:
    """Helper method for starting a :class:`TargetCli` or :class:`TargetHubCli` for one or multiple targets."""
    if cli := create_cli(targets, cli_cls):
        if commands is not None:
            for command in commands:
                cli.onecmd(command)
            return
        run_cli(cli)


def python_shell(targets: list[Target], commands: list[str] | None = None) -> None:
    """Helper method for starting a (I)Python shell with multiple targets."""
    banner = "Loaded targets in 'targets' variable. First target is in 't'."
    ns = {"targets": targets, "t": targets[0]}

    try:
        if commands is not None:
            for command in commands:
                eval(command, ns)
            return

        import IPython

        IPython.embed(header=banner, user_ns=ns, colors="linux")

        # IPython already prints an empty newline
    except ImportError:
        import code

        shell = code.InteractiveConsole(ns)
        shell.interact(banner)

        # Print an empty newline on exit
        print()


def create_cli(targets: list[Target], cli_cls: type[TargetCmd]) -> cmd.Cmd | None:
    """Helper method for instatiating the appropriate CLI."""
    if len(targets) == 1:
        target = targets[0]
        if not cli_cls.check_compatible(target):
            return None

        cli = cli_cls(target)
    else:
        cli = TargetHubCli(targets, cli_cls)

    return cli


def run_cli(cli: cmd.Cmd) -> None:
    """Helper method for running a cmd.Cmd cli.

    Loops cli.cmdloop(), skipping KeyboardInterrupts. This is done so
    that ctrl+c doesn't exit the shell but only resets the current line.
    """
    while True:
        try:
            cli.cmdloop()
        except KeyboardInterrupt:  # noqa: PERF203
            # Run postloop so the interrupted command is added to the history file
            cli.postloop()

            # Add a line when pressing ctrl+c, so the next one starts at a new line
            print()

        except Exception as e:
            if cli.debug:
                log.exception("Unhandled error")
            else:
                log.info(e)
                print(f"*** Unhandled error: {e}")
                print("If you wish to see the full debug trace, enable debug mode.")

            cli.postloop()
        else:
            # Print an empty newline on exit
            print()
            return


@catch_sigpipe
def main() -> int:
    help_formatter = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(
        description="dissect.target",
        fromfile_prefix_chars="@",
        formatter_class=help_formatter,
    )
    parser.add_argument("targets", metavar="TARGETS", nargs="*", help="targets to load")
    parser.add_argument("-p", "--python", action="store_true", help="(I)Python shell")
    parser.add_argument("-r", "--registry", action="store_true", help="registry shell")
    parser.add_argument("-c", "--commands", action="store", nargs="*", help="commands to execute")
    configure_generic_arguments(parser)

    args, _ = parser.parse_known_args()
    process_generic_arguments(parser, args)

    # For the shell tool we want -q to log slightly more then just CRITICAL messages.
    if args.quiet:
        logging.getLogger("dissect").setLevel(level=logging.ERROR)

    # PyPy < 3.10.14 readline is stuck in Python 2.7
    if platform.python_implementation() == "PyPy":
        major, minor, patch = tuple(map(int, platform.python_version_tuple()))
        if major < 3 or (major == 3 and (minor < 10 or (minor == 10 and patch < 14))):
            print(
                "Note for users of PyPy < 3.10.14:\n"
                "Autocomplete might not work due to an outdated version of pyrepl/readline.py\n"
                "To fix this, please update your version of PyPy.",
                file=sys.stderr,
            )

    try:
        open_shell(list(open_targets(args)), args.python, args.registry, args.commands)
    except TargetError as e:
        log.error("Error opening shell: %s", e)  # noqa: TRY400
        log.debug("", exc_info=e)

    return 0


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from flow.record.fieldtypes import datetime as dt

from dissect.target.plugins.apps.browser.chromium import ChromiumPlugin, decrypt_v10_linux
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


@pytest.fixture
def target_chromium_win(target_win_users: Target, fs_win: VirtualFilesystem) -> Target:
    fs_win.map_dir(
        "Users\\John\\AppData\\Local\\Chromium\\User Data\\Default\\",
        absolute_path("_data/plugins/apps/browser/chromium/"),
    )

    target_win_users.add_plugin(ChromiumPlugin)

    return target_win_users


@pytest.fixture
def target_chromium_unix(target_unix_users: Target, fs_unix: VirtualFilesystem) -> Target:
    fs_unix.map_dir("/root/.config/chromium/Default/", absolute_path("_data/plugins/apps/browser/chromium/"))
    target_unix_users.add_plugin(ChromiumPlugin)

    return target_unix_users


@pytest.mark.parametrize(
    "target_platform",
    ["target_chromium_win", "target_chromium_unix"],
)
def test_chromium_history(target_platform: Target, request: pytest.FixtureRequest) -> None:
    target_platform = request.getfixturevalue(target_platform)
    records = list(target_platform.chromium.history())

    assert len(records) == 5
    assert {"chromium"} == {record.browser for record in records}

    assert (
        records[0].url
        == "https://www.google.com/search?q=fox-it+github+dissect&oq=fox-it+github+dissect&gs_lcrp=EgZjaHJvbWUyBggA"
        "EEUYOTIHCAEQIRigAdIBCDU2OTNqMGo3qAIAsAIA&sourceid=chrome&ie=UTF-8"
    )
    assert records[0].id == 1
    assert records[0].visit_count == 2
    assert records[0].ts == dt("2022-12-22T12:14:26.396332+00:00")
    assert not records[0].hidden


@pytest.mark.parametrize(
    "target_platform",
    ["target_chromium_win", "target_chromium_unix"],
)
def test_chromium_downloads(target_platform: Target, request: pytest.FixtureRequest) -> None:
    target_platform = request.getfixturevalue(target_platform)
    records = list(target_platform.chromium.downloads())

    assert len(records) == 1
    assert {"chromium"} == {record.browser for record in records}

    assert records[0].id == 1
    assert records[0].ts_start == dt("2022-12-22T12:14:38.440832+00:00")
    assert records[0].ts_end == dt("2022-12-22T12:14:38.964170+00:00")
    assert records[0].url == "https://codeload.github.com/fox-it/dissect/zip/refs/heads/main"
    assert records[0].state == "complete"


@pytest.mark.parametrize(
    "target_platform",
    ["target_chromium_win", "target_chromium_unix"],
)
def test_chromium_cookies(target_platform: Target, request: pytest.FixtureRequest) -> None:
    target_platform = request.getfixturevalue(target_platform)
    records = list(target_platform.chromium.cookies())

    assert len(records) == 5
    assert {"chromium"} == {record.browser for record in records}

    assert sorted([*(c.name for c in records)]) == [
        "pl",
        "ssa-did",
        "ssa-sid",
        "tbb",
        "twk-theme",
    ]


@pytest.mark.parametrize(
    "target_platform",
    ["target_chromium_win", "target_chromium_unix"],
)
def test_chromium_extensions(target_platform: Target, request: pytest.FixtureRequest) -> None:
    target_platform = request.getfixturevalue(target_platform)
    records = list(target_platform.chromium.extensions())

    assert len(records) == 4
    assert {"chromium"} == {record.browser for record in records}

    assert records[0].ts_install == dt("2023-04-18T08:43:37.773874+00:00")
    assert records[0].ts_update == dt("2023-04-18T08:43:37.773874+00:00")
    assert records[0].name == "Web Store"
    assert records[0].version == "0.2"
    assert records[0].extension_id == "ahfgeienlihckogmohjhadlkjgocpleb"


def test_windows_chromium_passwords(target_chromium_win: Target) -> None:
    records = list(target_chromium_win.chromium.passwords())

    assert len(records) == 2

    for record in records:
        assert record.browser == "chromium"
        assert record.decrypted_username == "username"
        assert record.decrypted_password is None

    assert records[0].url == "https://example.com/"
    assert records[1].url == "https://example.org/"


def test_unix_chromium_passwords_basic(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_dir("/root/.config/chromium/Default/", absolute_path("_data/plugins/apps/browser/chromium/unix/basic/"))
    target_unix_users.add_plugin(ChromiumPlugin)

    records = list(target_unix_users.chromium.passwords())

    assert len(records) == 2

    for record in records:
        assert record.browser == "chromium"
        assert record.decrypted_username == "username"
        assert record.decrypted_password == "password"

    assert records[0].url == "https://example.com/"
    assert records[1].url == "https://example.org/"


def test_unix_chromium_passwords_gnome(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_dir("/root/.config/chromium/Default/", absolute_path("_data/plugins/apps/browser/chromium/unix/gnome/"))
    target_unix_users.add_plugin(ChromiumPlugin)

    records = list(target_unix_users.chromium.passwords())

    assert len(records) == 1

    assert records[0].decrypted_username == "username"
    assert records[0].decrypted_password is None
    assert records[0].url == "https://test.com/"


def test_decrypt_v10_linux_peanuts_key() -> None:
    """Test if we can decrypt a Linux V10 peanuts ciphertext."""
    encrypted = bytes.fromhex("763130d02645bb85e75ffdf893902f087b27a9")
    decrypted = decrypt_v10_linux(None, None, None, encrypted)
    assert decrypted == b"password"


def test_decrypt_v10_linux_empty_key() -> None:
    """Test if we can decrypt a Linux V10 empty key ciphertext."""
    encrypted = bytes.fromhex("763130195b29422c335652fb7a909368cbb3c6")
    decrypted = decrypt_v10_linux(None, None, None, encrypted)
    assert decrypted == b"password"



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING

from dissect.target.plugins.os.windows.activitiescache import ActivitiesCachePlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_windows_activitiescache_10_22H2(target_win_users: Target, fs_win: VirtualFilesystem) -> None:
    """Test if we can parse an ActivitiesCache.db file from Windows 10 22H2 correctly."""
    fs_win.map_file(
        "Users/John/AppData/Local/ConnectedDevicesPlatform/L.John/ActivitiesCache.db",
        str(absolute_path("_data/plugins/os/windows/activitiescache/ActivitiesCache.db")),
    )

    target_win_users.add_plugin(ActivitiesCachePlugin)

    records = list(target_win_users.activitiescache())

    assert len(records) == 22

    assert records[0].start_time == datetime(2025, 3, 4, 10, 28, 20, tzinfo=timezone.utc)
    assert not records[0].end_time
    assert records[0].last_modified_time == datetime(2025, 3, 4, 10, 28, 20, tzinfo=timezone.utc)
    assert records[0].last_modified_on_client == datetime(2025, 3, 4, 10, 28, 20, tzinfo=timezone.utc)
    assert not records[0].original_last_modified_on_client
    assert records[0].expiration_time == datetime(2032, 3, 2, 10, 28, 20, tzinfo=timezone.utc)
    assert records[0].activity_id == "c3d8d9afdec8125895bcc9348387ade1"
    assert (
        records[0].app_id
        == '[{"application":"microsoft.default.default","platform":"data_boundary"},{"application":"","platform":"packageId"},{"application":"","platform":"alternateId"}]'  # noqa: E501
    )
    assert not records[0].enterprise_id
    assert (
        records[0].app_activity_id
        == "default$windows.data.bluelightreduction.settings|windows.data.bluelightreduction.settings"
    )
    assert not records[0].group_app_activity_id
    assert records[0].group == "default$windows.data.bluelightreduction.settings"
    assert records[0].activity_type == 11
    assert records[0].activity_status == 1
    assert records[0].activity_priority == 3
    assert records[0].match_id == "default$windows.data.consent.consentitem$consents"
    assert records[0].etag == 1
    assert records[0].tag == "windows.data.bluelightreduction.settings"
    assert records[0].is_local_only
    assert records[0].created_in_cloud == datetime(1970, 1, 1, 0, 0, 0, tzinfo=timezone.utc)
    assert records[0].platform_device_id == "TNlommUcFlcw1lY+F+z4Sc57ss7UZKgX9GKxBTcbSYE="
    assert records[0].package_id_hash == "ZZzI03ZhPBD1Dfj2GvrCeHMuMjZ9FnVlYIDK+kZzTu0="
    assert records[0].payload == "Q0IBAMoUDhUAyh4OBwDKMgDKPAAA"
    assert not records[0].original_payload
    assert not records[0].clipboard_payload
    assert records[0].source == "C:\\Users\\John\\AppData\\Local\\ConnectedDevicesPlatform\\L.John\\ActivitiesCache.db"
    assert records[0].username == "John"



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest

from dissect.target.loader import open as loader_open
from dissect.target.loaders.vmwarevm import VmwarevmLoader
from dissect.target.target import Target

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator
    from pathlib import Path


@pytest.fixture
def mock_vmwarevm_dir(tmp_path: Path) -> Iterator[Path]:
    path = tmp_path / "Test.vmwarevm"
    path.mkdir()
    (path / "Test.vmx").touch()
    (path / "mock.vmdk").touch()

    with patch("dissect.hypervisor.descriptor.vmx.VMX") as mock_vmx:
        mock_vmx.parse.return_value = mock_vmx
        mock_vmx.disks.return_value = ["mock.vmdk"]

        yield path


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target], mock_vmwarevm_dir: Path) -> None:
    """Test that we correctly use ``VmwarevmLoader`` when opening a ``Target``."""
    with patch("dissect.target.container.open"), patch("dissect.target.target.Target.apply"):
        target = opener(mock_vmwarevm_dir)
        assert isinstance(target._loader, VmwarevmLoader)
        assert target.path == mock_vmwarevm_dir


def test_loader(mock_vmwarevm_dir: Path) -> None:
    """Test that ``VmwarevmLoader`` correctly loads a VMware VM directory."""
    loader = loader_open(mock_vmwarevm_dir)
    assert isinstance(loader, VmwarevmLoader)

    with patch("dissect.target.container.open") as mock_container_open:
        t = Target()
        loader.map(t)

        assert len(t.disks) == 1
        mock_container_open.assert_called_with(mock_vmwarevm_dir / "mock.vmdk")



from __future__ import annotations

import stat
from typing import TYPE_CHECKING
from unittest.mock import Mock

from dissect.target.filesystem import VirtualFile, VirtualFilesystem
from dissect.target.helpers import fsutil
from dissect.target.plugins.filesystem.unix.suid import SuidPlugin

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_suid_plugin(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    vfile = VirtualFile(fs_unix, "binary", None)
    vfile.lstat = Mock()
    vfile.lstat.return_value = fsutil.stat_result([stat.S_IFREG | stat.S_ISUID, 0, 0, 0, 0, 0, 0, 0, 0, 0])
    fs_unix.map_file_entry("/path/to/suid/binary", vfile)

    target_unix.add_plugin(SuidPlugin)

    results = list(target_unix.suid_binaries())
    assert len(results) == 1

    assert results[0].path == "/path/to/suid/binary"



"""Multilingual User Interface (MUI) translations for Windows en-US timezone strings.

This dictionary was exported from a ``tzres.dll.mui`` file from Windows Server 2019.

References:
    - Reversing DLL String Table in ``C:/Windows/System32/en-US/tzres.dll.mui``.
    - https://learn.microsoft.com/en-us/windows/win32/intl/using-registry-string-redirection
    - https://learn.microsoft.com/en-us/windows/win32/api/winnls/ns-winnls-filemuiinfo
"""

MUI_TZ_MAP = {
    10: "(UTC-01:00) Azores",
    11: "Azores Daylight Time",
    12: "Azores Standard Time",
    20: "(UTC-01:00) Cape Verde Is.",
    21: "Cape Verde Daylight Time",
    22: "Cape Verde Standard Time",
    30: "(UTC-02:00) Mid-Atlantic",
    31: "Mid-Atlantic Daylight Time",
    32: "Mid-Atlantic Standard Time",
    40: "(UTC-03:00) Brasilia",
    41: "E. South America Daylight Time",
    42: "E. South America Standard Time",
    50: "(UTC-03:00) Greenland",
    51: "Greenland Daylight Time",
    52: "Greenland Standard Time",
    60: "(UTC-03:00) Buenos Aires, Georgetown",
    61: "SA Eastern Daylight Time",
    62: "SA Eastern Standard Time",
    70: "(UTC-03:30) Newfoundland",
    71: "Newfoundland Daylight Time",
    72: "Newfoundland Standard Time",
    80: "(UTC-04:00) Atlantic Time (Canada)",
    81: "Atlantic Daylight Time",
    82: "Atlantic Standard Time",
    90: "(UTC-04:00) Santiago",
    91: "Pacific SA Daylight Time",
    92: "Pacific SA Standard Time",
    100: "(UTC-04:00) Caracas, La Paz",
    101: "SA Western Daylight Time",
    102: "SA Western Standard Time",
    103: "(UTC-04:00) Manaus",
    104: "Central Brazilian Daylight Time",
    105: "Central Brazilian Standard Time",
    110: "(UTC-05:00) Eastern Time (US & Canada)",
    111: "Eastern Daylight Time",
    112: "Eastern Standard Time",
    120: "(UTC-05:00) Bogota, Lima, Quito, Rio Branco",
    121: "SA Pacific Daylight Time",
    122: "SA Pacific Standard Time",
    130: "(UTC-05:00) Indiana (East)",
    131: "US Eastern Daylight Time",
    132: "US Eastern Standard Time",
    140: "(UTC-06:00) Saskatchewan",
    141: "Canada Central Daylight Time",
    142: "Canada Central Standard Time",
    150: "(UTC-06:00) Central America",
    151: "Central America Daylight Time",
    152: "Central America Standard Time",
    160: "(UTC-06:00) Central Time (US & Canada)",
    161: "Central Daylight Time",
    162: "Central Standard Time",
    170: "(UTC-06:00) Guadalajara, Mexico City, Monterrey",
    171: "Central Daylight Time (Mexico)",
    172: "Central Standard Time (Mexico)",
    180: "(UTC-07:00) Chihuahua, La Paz, Mazatlan",
    181: "Mountain Daylight Time (Mexico)",
    182: "Mountain Standard Time (Mexico)",
    190: "(UTC-07:00) Mountain Time (US & Canada)",
    191: "Mountain Daylight Time",
    192: "Mountain Standard Time",
    200: "(UTC-07:00) Arizona",
    201: "US Mountain Daylight Time",
    202: "US Mountain Standard Time",
    210: "(UTC-08:00) Pacific Time (US & Canada)",
    211: "Pacific Daylight Time",
    212: "Pacific Standard Time",
    213: "(UTC-08:00) Tijuana, Baja California",
    214: "Pacific Daylight Time (Mexico)",
    215: "Pacific Standard Time (Mexico)",
    220: "(UTC-09:00) Alaska",
    221: "Alaskan Daylight Time",
    222: "Alaskan Standard Time",
    230: "(UTC-10:00) Hawaii",
    231: "Hawaiian Daylight Time",
    232: "Hawaiian Standard Time",
    240: "(UTC-11:00) Midway Island, Samoa",
    241: "Samoa Daylight Time",
    242: "Samoa Standard Time",
    250: "(UTC-12:00) International Date Line West",
    251: "Dateline Daylight Time",
    252: "Dateline Standard Time",
    260: "(UTC) Dublin, Edinburgh, Lisbon, London",
    261: "GMT Daylight Time",
    262: "GMT Standard Time",
    270: "(UTC) Casablanca, Monrovia, Reykjavik",
    271: "Greenwich Daylight Time",
    272: "Greenwich Standard Time",
    280: "(UTC+01:00) Belgrade, Bratislava, Budapest, Ljubljana, Prague",
    281: "Central Europe Daylight Time",
    282: "Central Europe Standard Time",
    290: "(UTC+01:00) Sarajevo, Skopje, Warsaw, Zagreb",
    291: "Central European Daylight Time",
    292: "Central European Standard Time",
    300: "(UTC+01:00) Brussels, Copenhagen, Madrid, Paris",
    301: "Romance Daylight Time",
    302: "Romance Standard Time",
    310: "(UTC+01:00) West Central Africa",
    311: "W. Central Africa Daylight Time",
    312: "W. Central Africa Standard Time",
    320: "(UTC+01:00) Amsterdam, Berlin, Bern, Rome, Stockholm, Vienna",
    321: "W. Europe Daylight Time",
    322: "W. Europe Standard Time",
    330: "(UTC+02:00) Minsk",
    331: "E. Europe Daylight Time",
    332: "E. Europe Standard Time",
    333: "(UTC+02:00) Amman",
    334: "Jordan Daylight Time",
    335: "Jordan Standard Time",
    340: "(UTC+02:00) Cairo",
    341: "Egypt Daylight Time",
    342: "Egypt Standard Time",
    350: "(UTC+02:00) Helsinki, Kyiv, Riga, Sofia, Tallinn, Vilnius",
    351: "FLE Daylight Time",
    352: "FLE Standard Time",
    360: "(UTC+02:00) Athens, Bucharest, Istanbul",
    361: "GTB Daylight Time",
    362: "GTB Standard Time",
    363: "(UTC+02:00) Beirut",
    364: "Middle East Daylight Time",
    365: "Middle East Standard Time",
    370: "(UTC+02:00) Jerusalem",
    371: "Jerusalem Daylight Time",
    372: "Jerusalem Standard Time",
    380: "(UTC+02:00) Harare, Pretoria",
    381: "South Africa Daylight Time",
    382: "South Africa Standard Time",
    383: "(UTC+02:00) Windhoek",
    384: "Namibia Daylight Time",
    385: "Namibia Standard Time",
    390: "(UTC+03:00) Kuwait, Riyadh",
    391: "Arab Daylight Time",
    392: "Arab Standard Time",
    400: "(UTC+03:00) Baghdad",
    401: "Arabic Daylight Time",
    402: "Arabic Standard Time",
    410: "(UTC+03:00) Nairobi",
    411: "E. Africa Daylight Time",
    412: "E. Africa Standard Time",
    420: "(UTC+03:00) Moscow, St. Petersburg, Volgograd",
    421: "Russian Daylight Time",
    422: "Russian Standard Time",
    430: "(UTC+03:30) Tehran",
    431: "Iran Daylight Time",
    432: "Iran Standard Time",
    433: "(UTC+03:00) Tbilisi",
    434: "Georgian Daylight Time",
    435: "Georgian Standard Time",
    440: "(UTC+04:00) Abu Dhabi, Muscat",
    441: "Arabian Daylight Time",
    442: "Arabian Standard Time",
    447: "(UTC+04:00) Baku",
    448: "Azerbaijan Daylight Time",
    449: "Azerbaijan Standard Time",
    450: "(UTC+04:00) Yerevan",
    451: "Caucasus Daylight Time",
    452: "Caucasus Standard Time",
    460: "(UTC+04:30) Kabul",
    461: "Afghanistan Daylight Time",
    462: "Afghanistan Standard Time",
    470: "(UTC+05:00) Ekaterinburg",
    471: "Ekaterinburg Daylight Time",
    472: "Ekaterinburg Standard Time",
    480: "(UTC+05:00) Islamabad, Karachi, Tashkent",
    481: "West Asia Daylight Time",
    482: "West Asia Standard Time",
    490: "(UTC+05:30) Chennai, Kolkata, Mumbai, New Delhi",
    491: "India Daylight Time",
    492: "India Standard Time",
    500: "(UTC+05:45) Kathmandu",
    501: "Nepal Daylight Time",
    502: "Nepal Standard Time",
    510: "(UTC+06:00) Astana, Dhaka",
    511: "Central Asia Daylight Time",
    512: "Central Asia Standard Time",
    520: "(UTC+06:00) Almaty, Novosibirsk",
    521: "N. Central Asia Daylight Time",
    522: "N. Central Asia Standard Time",
    530: "(UTC+05:30) Sri Jayawardenepura",
    531: "Sri Lanka Daylight Time",
    532: "Sri Lanka Standard Time",
    540: "(UTC+06:30) Yangon (Rangoon)",
    541: "Myanmar Daylight Time",
    542: "Myanmar Standard Time",
    550: "(UTC+07:00) Krasnoyarsk",
    551: "North Asia Daylight Time",
    552: "North Asia Standard Time",
    560: "(UTC+07:00) Bangkok, Hanoi, Jakarta",
    561: "SE Asia Daylight Time",
    562: "SE Asia Standard Time",
    570: "(UTC+08:00) Beijing, Chongqing, Hong Kong, Urumqi",
    571: "China Daylight Time",
    572: "China Standard Time",
    580: "(UTC+08:00) Irkutsk, Ulaan Bataar",
    581: "North Asia East Daylight Time",
    582: "North Asia East Standard Time",
    590: "(UTC+08:00) Kuala Lumpur, Singapore",
    591: "Malay Peninsula Daylight Time",
    592: "Malay Peninsula Standard Time",
    600: "(UTC+08:00) Taipei",
    601: "Taipei Daylight Time",
    602: "Taipei Standard Time",
    610: "(UTC+08:00) Perth",
    611: "W. Australia Daylight Time",
    612: "W. Australia Standard Time",
    620: "(UTC+09:00) Seoul",
    621: "Korea Daylight Time",
    622: "Korea Standard Time",
    630: "(UTC+09:00) Osaka, Sapporo, Tokyo",
    631: "Tokyo Daylight Time",
    632: "Tokyo Standard Time",
    640: "(UTC+09:00) Yakutsk",
    641: "Yakutsk Daylight Time",
    642: "Yakutsk Standard Time",
    650: "(UTC+09:30) Darwin",
    651: "AUS Central Daylight Time",
    652: "AUS Central Standard Time",
    660: "(UTC+09:30) Adelaide",
    661: "Cen. Australia Daylight Time",
    662: "Cen. Australia Standard Time",
    670: "(UTC+10:00) Canberra, Melbourne, Sydney",
    671: "AUS Eastern Daylight Time",
    672: "AUS Eastern Standard Time",
    680: "(UTC+10:00) Brisbane",
    681: "E. Australia Daylight Time",
    682: "E. Australia Standard Time",
    690: "(UTC+10:00) Hobart",
    691: "Tasmania Daylight Time",
    692: "Tasmania Standard Time",
    700: "(UTC+10:00) Vladivostok",
    701: "Vladivostok Daylight Time",
    702: "Vladivostok Standard Time",
    710: "(UTC+10:00) Guam, Port Moresby",
    711: "West Pacific Daylight Time",
    712: "West Pacific Standard Time",
    720: "(UTC+11:00) Magadan, Solomon Is., New Caledonia",
    721: "Central Pacific Daylight Time",
    722: "Central Pacific Standard Time",
    730: "(UTC+12:00) Fiji, Kamchatka, Marshall Is.",
    731: "Fiji Daylight Time",
    732: "Fiji Standard Time",
    740: "(UTC+12:00) Auckland, Wellington",
    741: "New Zealand Daylight Time",
    742: "New Zealand Standard Time",
    750: "(UTC+13:00) Nuku'alofa",
    751: "Tonga Daylight Time",
    752: "Tonga Standard Time",
    770: "(UTC-03:00) Montevideo",
    771: "Montevideo Daylight Time",
    772: "Montevideo Standard Time",
    790: "(UTC-04:00) La Paz",
    791: "SA Western Daylight Time",
    792: "SA Western Standard Time",
    810: "(UTC-04:30) Caracas",
    811: "Venezuela Daylight Time",
    812: "Venezuela Standard Time",
    830: "(UTC-03:00) Georgetown",
    831: "SA Eastern Daylight Time",
    832: "SA Eastern Standard Time",
    840: "(UTC-03:00) Buenos Aires",
    841: "Argentina Daylight Time",
    842: "Argentina Standard Time",
    860: "(UTC+05:00) Tashkent",
    870: "(UTC+05:00) Islamabad, Karachi",
    871: "Pakistan Daylight Time",
    872: "Pakistan Standard Time",
    880: "(UTC) Monrovia, Reykjavik",
    890: "(UTC) Casablanca",
    891: "Morocco Daylight Time",
    892: "Morocco Standard Time",
    910: "(UTC+04:00) Port Louis",
    911: "Mauritius Daylight Time",
    912: "Mauritius Standard Time",
    930: "(UTC) Coordinated Universal Time",
    931: "Coordinated Universal Time",
    932: "Coordinated Universal Time",
    940: "(UTC-03:00) Cayenne",
    950: "(UTC-04:00) Georgetown, La Paz, San Juan",
    960: "(UTC-04:00) Asuncion",
    961: "Paraguay Daylight Time",
    962: "Paraguay Standard Time",
    970: "(UTC+12:00) Fiji, Marshall Is.",
    980: "(UTC+12:00) Petropavlovsk-Kamchatsky",
    981: "Kamchatka Daylight Time",
    982: "Kamchatka Standard Time",
    990: "(UTC-05:00) Bogota, Lima, Quito",
    1010: "(UTC+06:00) Astana",
    1020: "(UTC+06:00) Dhaka",
    1021: "Bangladesh Daylight Time",
    1022: "Bangladesh Standard Time",
    1030: "(UTC+08:00) Irkutsk",
    1040: "(UTC+08:00) Ulaanbaatar",
    1041: "Ulaanbaatar Daylight Time",
    1042: "Ulaanbaatar Standard Time",
    1050: "(UTC-11:00) Samoa",
    1060: "(UTC-11:00) Midway Islands",
    1061: "North Pacific Daylight Time",
    1062: "North Pacific Standard Time",
    1070: "(UTC+04:00) Tbilisi",
    1080: "(UTC+06:00) Novosibirsk",
    1100: "(UTC-08:00) Baja California",
    1110: "(UTC-03:00) Cayenne, Fortaleza",
    1120: "(UTC-04:00) Cuiaba",
    1130: "(UTC-04:00) Georgetown, La Paz, Manaus, San Juan",
    1140: "(UTC+12:00) Fiji",
    1150: "(UTC-01:00) Coordinated Universal Time-01",
    1151: "UTC-01",
    1152: "UTC-01",
    1160: "(UTC-02:00) Coordinated Universal Time-02",
    1161: "UTC-02",
    1162: "UTC-02",
    1170: "(UTC-03:00) Coordinated Universal Time-03",
    1171: "UTC-03",
    1172: "UTC-03",
    1180: "(UTC-04:00) Coordinated Universal Time-04",
    1181: "UTC-04",
    1182: "UTC-04",
    1190: "(UTC-05:00) Coordinated Universal Time-05",
    1191: "UTC-05",
    1192: "UTC-05",
    1200: "(UTC-06:00) Coordinated Universal Time-06",
    1201: "UTC-06",
    1202: "UTC-06",
    1210: "(UTC-07:00) Coordinated Universal Time-07",
    1211: "UTC-07",
    1212: "UTC-07",
    1220: "(UTC-08:00) Coordinated Universal Time-08",
    1221: "UTC-08",
    1222: "UTC-08",
    1230: "(UTC-09:00) Coordinated Universal Time-09",
    1231: "UTC-09",
    1232: "UTC-09",
    1240: "(UTC-10:00) Coordinated Universal Time-10",
    1241: "UTC-10",
    1242: "UTC-10",
    1250: "(UTC-11:00) Coordinated Universal Time-11",
    1251: "UTC-11",
    1252: "UTC-11",
    1260: "(UTC-12:00) Coordinated Universal Time-12",
    1261: "UTC-12",
    1262: "UTC-12",
    1270: "(UTC+01:00) Coordinated Universal Time+01",
    1271: "UTC+01",
    1272: "UTC+01",
    1280: "(UTC+02:00) Coordinated Universal Time+02",
    1281: "UTC+02",
    1282: "UTC+02",
    1290: "(UTC+03:00) Coordinated Universal Time+03",
    1291: "UTC+03",
    1292: "UTC+03",
    1300: "(UTC+04:00) Coordinated Universal Time+04",
    1301: "UTC+04",
    1302: "UTC+04",
    1310: "(UTC+05:00) Coordinated Universal Time+05",
    1311: "UTC+05",
    1312: "UTC+05",
    1320: "(UTC+06:00) Coordinated Universal Time+06",
    1321: "UTC+06",
    1322: "UTC+06",
    1330: "(UTC+07:00) Coordinated Universal Time+07",
    1331: "UTC+07",
    1332: "UTC+07",
    1340: "(UTC+08:00) Coordinated Universal Time+08",
    1341: "UTC+08",
    1342: "UTC+08",
    1350: "(UTC+09:00) Coordinated Universal Time+09",
    1351: "UTC+09",
    1352: "UTC+09",
    1360: "(UTC+10:00) Coordinated Universal Time+10",
    1361: "UTC+10",
    1362: "UTC+10",
    1370: "(UTC+11:00) Coordinated Universal Time+11",
    1371: "UTC+11",
    1372: "UTC+11",
    1380: "(UTC+12:00) Coordinated Universal Time+12",
    1381: "UTC+12",
    1382: "UTC+12",
    1390: "(UTC+13:00) Coordinated Universal Time+13",
    1391: "UTC+13",
    1392: "UTC+13",
    1410: "(UTC+02:00) Damascus",
    1411: "Syria Daylight Time",
    1412: "Syria Standard Time",
    1420: "(UTC+12:00) Petropavlovsk-Kamchatsky - Old",
    1440: "(UTC+01:00) Windhoek",
    1460: "(UTC+11:00) Solomon Is., New Caledonia",
    1470: "(UTC+11:00) Magadan",
    1471: "Magadan Daylight Time",
    1472: "Magadan Standard Time",
    1490: "(UTC+02:00) Athens, Bucharest",
    1500: "(UTC+02:00) Istanbul",
    1501: "Turkey Daylight Time",
    1502: "Turkey Standard Time",
    1520: "(UTC+04:00) Moscow, St. Petersburg, Volgograd",
    1530: "(UTC+06:00) Ekaterinburg",
    1540: "(UTC+07:00) Novosibirsk",
    1550: "(UTC+08:00) Krasnoyarsk",
    1560: "(UTC+09:00) Irkutsk",
    1570: "(UTC+10:00) Yakutsk",
    1580: "(UTC+11:00) Vladivostok",
    1590: "(UTC+12:00) Magadan",
    1600: "(UTC+03:00) Kaliningrad",
    1601: "Kaliningrad Daylight Time",
    1602: "Kaliningrad Standard Time",
    1620: "(UTC+03:00) Kaliningrad, Minsk",
    1630: "(UTC+02:00) Nicosia",
    1640: "(UTC+13:00) Samoa",
    1660: "(UTC-03:00) Salvador",
    1661: "Bahia Daylight Time",
    1662: "Bahia Standard Time",
    1680: "(UTC+02:00) E. Europe",
    1700: "(UTC+03:00) Amman",
    1720: "(UTC+02:00) Tripoli",
    1721: "Libya Daylight Time",
    1722: "Libya Standard Time",
    1740: "(UTC+05:00) Ashgabat, Tashkent",
    1760: "(UTC-02:00) Mid-Atlantic - Old",
    1780: "(UTC+02:00) Tripoli",
    1800: "(UTC+14:00) Kiritimati Island",
    1801: "Line Islands Daylight Time",
    1802: "Line Islands Standard Time",
    1820: "(UTC+02:00) Kaliningrad (RTZ 1)",
    1821: "Russia TZ 1 Daylight Time",
    1822: "Russia TZ 1 Standard Time",
    1830: "(UTC+03:00) Moscow, St. Petersburg, Volgograd (RTZ 2)",
    1831: "Russia TZ 2 Daylight Time",
    1832: "Russia TZ 2 Standard Time",
    1840: "(UTC+05:00) Ekaterinburg (RTZ 4)",
    1841: "Russia TZ 4 Daylight Time",
    1842: "Russia TZ 4 Standard Time",
    1850: "(UTC+06:00) Novosibirsk (RTZ 5)",
    1851: "Russia TZ 5 Daylight Time",
    1852: "Russia TZ 5 Standard Time",
    1860: "(UTC+07:00) Krasnoyarsk (RTZ 6)",
    1861: "Russia TZ 6 Daylight Time",
    1862: "Russia TZ 6 Standard Time",
    1870: "(UTC+08:00) Irkutsk (RTZ 7)",
    1871: "Russia TZ 7 Daylight Time",
    1872: "Russia TZ 7 Standard Time",
    1880: "(UTC+09:00) Yakutsk (RTZ 8)",
    1881: "Russia TZ 8 Daylight Time",
    1882: "Russia TZ 8 Standard Time",
    1890: "(UTC+04:00) Izhevsk, Samara (RTZ 3)",
    1891: "Russia TZ 3 Daylight Time",
    1892: "Russia TZ 3 Standard Time",
    1910: "(UTC+11:00) Chokurdakh (RTZ 10)",
    1911: "Russia TZ 10 Daylight Time",
    1912: "Russia TZ 10 Standard Time",
    1930: "(UTC+12:00) Anadyr, Petropavlovsk-Kamchatsky (RTZ 11)",
    1931: "Russia TZ 11 Daylight Time",
    1932: "Russia TZ 11 Standard Time",
    1950: "(UTC+10:00) Vladivostok, Magadan (RTZ 9)",
    1951: "Russia TZ 9 Daylight Time",
    1952: "Russia TZ 9 Standard Time",
    1970: "(UTC+03:00) Minsk",
    1971: "Belarus Daylight Time",
    1972: "Belarus Standard Time",
    1980: "(UTC+10:00) Magadan",
    2000: "(UTC-01:00) Cabo Verde Is.",
    2001: "Cabo Verde Daylight Time",
    2002: "Cabo Verde Standard Time",
    2020: "(UTC-03:00) Santiago",
    2040: "(UTC-05:00) Chetumal",
    2041: "Eastern Daylight Time (Mexico)",
    2042: "Eastern Standard Time (Mexico)",
    2060: "(UTC+08:30) Pyongyang",
    2061: "North Korea Daylight Time",
    2062: "North Korea Standard Time",
    2080: "(UTC-03:00) City of Buenos Aires",
    2120: "(UTC+10:00) Vladivostok",
    2140: "(UTC+09:00) Chita",
    2141: "Transbaikal Daylight Time",
    2142: "Transbaikal Standard Time",
    2160: "(UTC+07:00) Barnaul, Gorno-Altaysk",
    2161: "Altai Daylight Time",
    2162: "Altai Standard Time",
    2180: "(UTC+04:00) Astrakhan, Ulyanovsk",
    2181: "Astrakhan Daylight Time",
    2182: "Astrakhan Standard Time",
    2190: "(UTC+05:00) Ekaterinburg",
    2200: "(UTC+02:00) Kaliningrad",
    2220: "(UTC+06:00) Novosibirsk",
    2230: "(UTC+08:00) Irkutsk",
    2240: "(UTC+07:00) Krasnoyarsk",
    2250: "(UTC+11:00) Chokurdakh",
    2260: "(UTC+12:00) Anadyr, Petropavlovsk-Kamchatsky",
    2270: "(UTC+04:00) Izhevsk, Samara",
    2280: "(UTC+03:00) Moscow, St. Petersburg, Volgograd",
    2290: "(UTC+09:00) Yakutsk",
    2310: "(UTC+11:00) Magadan, Sakhalin",
    2320: "(UTC+11:00) Sakhalin",
    2321: "Sakhalin Daylight Time",
    2322: "Sakhalin Standard Time",
    2340: "(UTC-05:00) Haiti",
    2341: "Haiti Daylight Time",
    2342: "Haiti Standard Time",
    2370: "(UTC-06:00) Easter Island",
    2371: "Easter Island Daylight Time",
    2372: "Easter Island Standard Time",
    2390: "(UTC-10:00) Aleutian Islands",
    2391: "Aleutian Daylight Time",
    2392: "Aleutian Standard Time",
    2410: "(UTC-09:30) Marquesas Islands",
    2411: "Marquesas Daylight Time",
    2412: "Marquesas Standard Time",
    2430: "(UTC-05:00) Havana",
    2431: "Cuba Daylight Time",
    2432: "Cuba Standard Time",
    2450: "(UTC-03:00) Saint Pierre and Miquelon",
    2451: "Saint Pierre Daylight Time",
    2452: "Saint Pierre Standard Time",
    2470: "(UTC+02:00) Gaza, Hebron",
    2471: "West Bank Gaza Daylight Time",
    2472: "West Bank Gaza Standard Time",
    2490: "(UTC+08:45) Eucla",
    2491: "Aus Central W. Daylight Time",
    2492: "Aus Central W. Standard Time",
    2510: "(UTC+10:30) Lord Howe Island",
    2511: "Lord Howe Daylight Time",
    2512: "Lord Howe Standard Time",
    2530: "(UTC+12:45) Chatham Islands",
    2531: "Chatham Islands Daylight Time",
    2532: "Chatham Islands Standard Time",
    2550: "(UTC+07:00) Hovd",
    2551: "W. Mongolia Daylight Time",
    2552: "W. Mongolia Standard Time",
    2570: "(UTC-04:00) Turks and Caicos",
    2571: "Turks and Caicos Daylight Time",
    2572: "Turks and Caicos Standard Time",
    2590: "(UTC-03:00) Araguaina",
    2591: "Tocantins Daylight Time",
    2592: "Tocantins Standard Time",
    2610: "(UTC+11:00) Bougainville Island",
    2611: "Bougainville Daylight Time",
    2612: "Bougainville Standard Time",
    2630: "(UTC+11:00) Norfolk Island",
    2631: "Norfolk Daylight Time",
    2632: "Norfolk Standard Time",
    2650: "(UTC+00:00) Casablanca",
    2670: "(UTC+00:00) Dublin, Edinburgh, Lisbon, London",
    2690: "(UTC+00:00) Monrovia, Reykjavik",
    2710: "(UTC+02:00) Chisinau",
    2730: "(UTC-04:00) Caracas",
    2750: "(UTC+07:00) Tomsk",
    2751: "Tomsk Daylight Time",
    2752: "Tomsk Standard Time",
    2770: "(UTC+06:00) Omsk",
    2771: "Omsk Daylight Time",
    2772: "Omsk Standard Time",
    2790: "(UTC+07:00) Novosibirsk",
    2791: "Novosibirsk Daylight Time",
    2792: "Novosibirsk Standard Time",
    2810: "(UTC+03:00) Istanbul",
    2840: "(UTC+04:00) Saratov",
    2841: "Saratov Daylight Time",
    2842: "Saratov Standard Time",
    2870: "(UTC-03:00) Punta Arenas",
    2871: "Magallanes Daylight Time",
    2872: "Magallanes Standard Time",
    2890: "(UTC+02:00) Khartoum",
    2891: "Sudan Daylight Time",
    2892: "Sudan Standard Time",
    2910: "(UTC-05:00) Turks and Caicos",
    2940: "(UTC+01:00) Sao Tome",
    2941: "Sao Tome Daylight Time",
    2942: "Sao Tome Standard Time",
    2960: "(UTC+09:00) Pyongyang",
    2980: "(UTC+03:00) Moscow, St. Petersburg",
    2990: "(UTC+04:00) Volgograd",
    2991: "Volgograd Daylight Time",
    2992: "Volgograd Standard Time",
    3010: "(UTC+01:00) Casablanca",
    3030: "(UTC+00:00) Sao Tome",
    3050: "(UTC+05:00) Qyzylorda",
    3051: "Qyzylorda Daylight Time",
    3052: "Qyzylorda Standard Time",
    3080: "(UTC-08:00) Yukon",
    3081: "Yukon Daylight Time",
    3082: "Yukon Standard Time",
    3090: "(UTC-07:00) Yukon",
    3110: "(UTC+03:00) Volgograd",
    3140: "(UTC+02:00) Juba",
    3141: "South Sudan Daylight Time",
    3142: "South Sudan Standard Time",
}



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.os.unix.linux.debian.dpkg import (
    STATUS_FILE_NAME,
    DpkgPlugin,
)
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_dpkg_plugin_status(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    status_file = absolute_path("_data/plugins/os/unix/linux/debian/dpkg/dpkg-status")

    fs_unix.map_file(STATUS_FILE_NAME, status_file)

    target_unix.add_plugin(DpkgPlugin)

    results = list(target_unix.dpkg.status())

    assert len(results) == 9
    assert {r.section for r in results} == {"admin", "utils", "libs", "libdevel"}
    assert {r.arch for r in results} == {"amd64", "all"}


def test_dpkg_plugin_log_plain(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    log_file = absolute_path("_data/plugins/os/unix/linux/debian/dpkg/dpkg.log")
    fs_unix.map_file("/var/log/dpkg.log", log_file)

    target_unix.add_plugin(DpkgPlugin)

    results = list(target_unix.dpkg.log())

    assert len(results) == 362
    assert all(r.name for r in results)
    # no remove operations in this test set
    assert all(r.version for r in results)
    assert {r.operation for r in results} == {"install", "upgrade", "status"}
    assert {r.arch for r in results} == {"amd64", "all"}


def test_dpkg_plugin_log_gz(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    log_file_gz = absolute_path("_data/plugins/os/unix/linux/debian/dpkg/dpkg.log.2.gz")
    fs_unix.map_file("/var/log/dpkg.log.2.gz", log_file_gz)

    target_unix.add_plugin(DpkgPlugin)

    results = list(target_unix.dpkg.log())

    assert len(results) == 97
    assert all(r.name for r in results)

    # remove operations are in the test set
    seen_versions = {r.version for r in results}
    assert None in seen_versions
    assert len(seen_versions) > 1

    assert {r.operation for r in results} == {"trigproc", "upgrade", "status", "remove"}
    assert {r.arch for r in results} == {"amd64", "all"}



from __future__ import annotations

import zipfile as zf
from typing import TYPE_CHECKING

from dissect.target.helpers.lazy import import_lazy
from dissect.target.loader import Loader, SubLoader
from dissect.target.loaders.dir import find_and_map_dirs

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import target


ZIP_EXT = (".zip",)


class ZipSubLoader(SubLoader[zf.Path]):
    """Zip implementation of a :class:`SubLoader`."""

    def __init__(self, path: Path, zipfile: zf.Path, **kwargs):
        super().__init__(path, zipfile, **kwargs)
        self.zip = zipfile

    @staticmethod
    def detect(path: Path, zipfile: zf.Path) -> bool:
        """Only to be called internally by :class:`ZipLoader`."""
        raise NotImplementedError

    def map(self, target: target.Target) -> None:
        """Only to be called internally by :class:`ZipLoader`."""
        raise NotImplementedError


class GenericZipSubLoader(ZipSubLoader):
    """Generic zip sub loader."""

    @staticmethod
    def detect(path: Path, zipfile: zf.Path) -> bool:
        return True

    def map(self, target: target.Target) -> None:
        find_and_map_dirs(target, self.zip)


class ZipLoader(Loader):
    """Load zip files."""

    __subloaders__ = (
        import_lazy("dissect.target.loaders.acquire").AcquireZipSubLoader,
        import_lazy("dissect.target.loaders.uac").UacZipSubLoader,
        GenericZipSubLoader,  # should be last
    )

    def __init__(self, path: Path, **kwargs):
        super().__init__(path, **kwargs)

        self.fh = path.open("rb")
        self.zip = zf.Path(self.fh)
        self.subloader = None

    @staticmethod
    def detect(path: Path) -> bool:
        return path.name.lower().endswith(ZIP_EXT)

    def map(self, target: target.Target) -> None:
        for candidate in self.__subloaders__:
            if candidate.detect(self.path, self.zip):
                self.subloader = candidate(self.path, self.zip, parsed_path=self.parsed_path)
                self.subloader.map(target)
                break



from __future__ import annotations

from datetime import datetime, timezone
from io import BytesIO
from textwrap import dedent
from typing import TYPE_CHECKING

from dissect.target.plugins.apps.remoteaccess.teamviewer import TeamViewerPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_teamviewer_global_log(target_win_users: Target, fs_win: VirtualFilesystem) -> None:
    teamviewer_logfile = absolute_path("_data/plugins/apps/remoteaccess/teamviewer/TestTeamviewer.log")
    target_logfile_name = "/sysvol/Program Files/TeamViewer/TestTeamviewer.log"

    _, _, map_path = target_logfile_name.partition("sysvol/")
    fs_win.map_file(map_path, teamviewer_logfile)

    tvp = TeamViewerPlugin(target_win_users)

    records = list(tvp.logs())
    assert len(records) == 4

    assert records[0].ts == datetime(2021, 11, 11, 12, 34, 56, tzinfo=timezone.utc)
    assert records[0].message == "Strip the headers, trace the source!"
    assert records[0].source == target_logfile_name
    assert records[0].username is None
    assert records[0].user_id is None
    assert records[0].user_home is None


def test_teamviewer_user_log(target_win_users: Target, fs_win: VirtualFilesystem) -> None:
    teamviewer_logfile = absolute_path("_data/plugins/apps/remoteaccess/teamviewer/TestTeamviewer.log")
    user_details = target_win_users.user_details.find(username="John")
    target_logfile_name = f"{user_details.home_path}/appdata/roaming/teamviewer/teamviewer_TEST_logfile.log"

    _, _, map_path = target_logfile_name.partition("C:/")
    fs_win.map_file(map_path, teamviewer_logfile)

    tvp = TeamViewerPlugin(target_win_users)

    records = list(tvp.logs())
    assert len(records) == 4

    assert records[0].ts == datetime(2021, 11, 11, 12, 34, 56, tzinfo=timezone.utc)
    assert records[0].message == "Strip the headers, trace the source!"
    assert records[0].source == target_logfile_name
    assert records[0].username == user_details.user.name
    assert records[0].user_id == user_details.user.sid
    assert records[0].user_home == user_details.user.home


def test_teamviewer_special_date_parsing(target_win_users: Target, fs_win: VirtualFilesystem) -> None:
    teamviewer_logfile = absolute_path("_data/plugins/apps/remoteaccess/teamviewer/TestTeamviewer.log")
    user_details = target_win_users.user_details.find(username="John")
    target_logfile_name = f"{user_details.home_path}/appdata/roaming/teamviewer/teamviewer_TEST_logfile.log"

    _, _, map_path = target_logfile_name.partition("C:/")
    fs_win.map_file(map_path, teamviewer_logfile)

    tvp = TeamViewerPlugin(target_win_users)

    records = list(tvp.logs())
    assert len(records) == 4

    assert records[1].ts == datetime(2021, 11, 11, 12, 35, 55, 465000, tzinfo=timezone.utc)
    assert records[1].message == "Should be year 2021"

    assert records[2].ts == datetime(2021, 11, 11, 12, 36, 11, 111000, tzinfo=timezone.utc)
    assert records[2].message == "Should discard the milliseconds properly"

    assert records[3].ts == datetime(2021, 11, 11, 12, 37, 0, 0, tzinfo=timezone.utc)
    assert records[3].message == "Should be year 2021"


def test_teamviewer_timezone(target_win_users: Target, fs_win: VirtualFilesystem) -> None:
    """Test if we correctly set the timezone in teamviewer logs."""

    log = """
    Start:          2024/12/31 01:02:03.123 (UTC+2:00)
    2024/12/31 01:02:03.200  1234  5678 G1   LanguageControl: device language is 'enUS'
    2024/12/31 01:02:03.300  1234  5678 G1   Example message 1
    2024/12/31 01:02:03.400  1234  5678 G1   Example message 2
    2024/12/31 01:02:03.500  1234  5678 G1   Example message 3
    2024/12/31 01:02:03.600  1234  5678 G1   Example message 4
    2024/12/31 01:02:03.700  1234  5678 G1!! Example message 5
    2024/12/31 01:02:03.800  1234  5678 G1   TeamViewer is going offline!
    2024/12/31 01:02:03.900  1234  5678 G1   NetworkControl shutdown done
    """
    fs_win.map_file_fh(
        "Users/John/AppData/Roaming/TeamViewer/TeamViewer1337_Logfile.log", BytesIO(dedent(log).encode())
    )

    target_win_users.add_plugin(TeamViewerPlugin)

    records = sorted(target_win_users.teamviewer.logs(), key=lambda r: r.ts)

    assert len(records) == 8

    # 01:02:03 with UTC+0200 becomes 23:02:03 UTC
    assert records[0].ts == datetime(2024, 12, 30, 23, 2, 3, 200000, tzinfo=timezone.utc)
    assert records[0].message == "1234  5678 G1   LanguageControl: device language is 'enUS'"
    assert records[0].source == "C:\\Users\\John\\AppData\\Roaming\\TeamViewer\\TeamViewer1337_Logfile.log"
    assert records[0].username == "John"


def test_teamviewer_incoming(target_win_users: Target, fs_win: VirtualFilesystem) -> None:
    """Test TeamViewer incoming connection log parsing."""
    log = """
    1031857653	DESKTOP-CAK7OMO	11-09-2022 14:44:03	11-09-2022 15:27:53	SERVER TV	RemoteControl	{C2CC2F16-D1F4-4547-9928-EE63891D4CC0}
    1031857653	DESKTOP-CAK7OMO	22-12-2022 19:25:22	22-12-2022 19:49:28	Server	RemoteControl	{4BF22BA7-32BA-4F64-8755-97E6E45F9883}
    """  # noqa: E501
    fs_win.map_file_fh("Program Files/TeamViewer/Connections_incoming.txt", BytesIO(dedent(log).encode()))

    target_win_users.add_plugin(TeamViewerPlugin)

    records = list(target_win_users.teamviewer.incoming())
    assert len(records) == 2

    assert records[0].ts == datetime(2022, 9, 11, 14, 44, 3, tzinfo=timezone.utc)
    assert records[0].end == datetime(2022, 9, 11, 15, 27, 53, tzinfo=timezone.utc)
    assert records[0].remote_id == "1031857653"
    assert records[0].name == "DESKTOP-CAK7OMO"
    assert records[0].user == "SERVER TV"
    assert records[0].connection_type == "RemoteControl"
    assert records[0].connection_id == "{C2CC2F16-D1F4-4547-9928-EE63891D4CC0}"

    assert records[1].ts == datetime(2022, 12, 22, 19, 25, 22, tzinfo=timezone.utc)
    assert records[1].end == datetime(2022, 12, 22, 19, 49, 28, tzinfo=timezone.utc)
    assert records[1].remote_id == "1031857653"
    assert records[1].name == "DESKTOP-CAK7OMO"
    assert records[1].user == "Server"
    assert records[1].connection_type == "RemoteControl"
    assert records[1].connection_id == "{4BF22BA7-32BA-4F64-8755-97E6E45F9883}"



from __future__ import annotations

from typing import TYPE_CHECKING

from defusedxml import ElementTree
from dissect.hypervisor.descriptor.vbox import NS

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import ChildTargetRecord
from dissect.target.plugin import ChildTargetPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.target import Target


class VirtualBoxChildTargetPlugin(ChildTargetPlugin):
    """Child target plugin that yields from Oracle VirtualBox VMs.

    Tested with configuration files from major versions 5, 6 and 7.

    References:
        - https://docs.oracle.com/en/virtualization/virtualbox/6.1/admin/TechnicalBackground.html
    """

    __type__ = "virtualbox"

    USER_PATHS = (
        # Windows
        ".VirtualBox",
        # Linux
        ".config/VirtualBox",
        # macOS
        "Library/VirtualBox",
    )

    DEFAULT_PATHS = ("VirtualBox VMs",)

    def __init__(self, target: Target):
        super().__init__(target)
        self.vboxes = list(self.find_vms())

    def find_vms(self) -> Iterator[Path]:
        """Yield Oracle VirtualBox ``.vbox`` file(s) found on the target."""
        seen = set()
        for user_details in self.target.user_details.all_with_home():
            # Yield `.vbox` from default locations and add to seen.
            for default_path in self.DEFAULT_PATHS:
                for path in user_details.home_path.joinpath(default_path).glob("*/*.vbox"):
                    if path not in seen and path.is_file():
                        seen.add(path)
                        yield path

            # Parse VirtualBox.xml configs
            for user_path in self.USER_PATHS:
                # Search for both VirtualBox.xml and VirtualBox.xml-prev
                for path in user_details.home_path.joinpath(user_path).glob("VirtualBox.xml*"):
                    if not path.is_file():
                        self.target.log.warning("Unable to parse %s: not a file", path)
                        continue

                    try:
                        config = ElementTree.fromstring(path.read_text())
                    except Exception as e:
                        self.target.log.warning("Unable to parse %s: %s", path, e)
                        self.target.log.debug("", exc_info=e)
                        continue

                    # Parse MachineEntries
                    for machine in config.findall(f".//{NS}MachineEntry"):
                        if (
                            (src := machine.get("src"))
                            and (src_path := self.target.fs.path(src)).exists()
                            and src_path not in seen
                        ):
                            seen.add(src_path)
                            yield src_path

                    # Glob for SystemProperties defaultMachineFolder
                    for system_properties in config.findall(f".//{NS}SystemProperties"):
                        if (folder_str := system_properties.get("defaultMachineFolder")) and (
                            folder_dir := self.target.fs.path(folder_str)
                        ).is_dir():
                            for vbox_file in folder_dir.glob("*/*.vbox"):
                                if vbox_file not in seen:
                                    seen.add(vbox_file)
                                    yield vbox_file

    def check_compatible(self) -> None:
        if not self.vboxes:
            raise UnsupportedPluginError("No VirtualBox children found on target")

    def list_children(self) -> Iterator[ChildTargetRecord]:
        for vbox in self.vboxes:
            try:
                config = ElementTree.fromstring(vbox.read_bytes())
                name = config.find(f".//{NS}Machine").attrib["name"]
            except Exception as e:
                self.target.log.error("Failed to parse name from VirtualBox XML: %s", vbox)  # noqa: TRY400
                self.target.log.debug("", exc_info=e)

                name = None

            yield ChildTargetRecord(
                type=self.__type__,
                name=name,
                path=vbox,
                _target=self.target,
            )



from __future__ import annotations

import logging
import re
from collections import defaultdict
from functools import lru_cache
from typing import TYPE_CHECKING, Final

from dissect.target.exceptions import (
    HiveUnavailableError,
    RegistryKeyNotFoundError,
    RegistryValueNotFoundError,
    UnsupportedPluginError,
)
from dissect.target.helpers.regutil import (
    HiveCollection,
    KeyCollection,
    RegfHive,
    RegistryHive,
    RegistryKey,
    RegistryValue,
    ValueCollection,
    VirtualHive,
    glob_ext,
    glob_split,
)
from dissect.target.plugin import Plugin, internal

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.helpers.record import WindowsUserRecord
    from dissect.target.plugins.general.users import UserDetails
    from dissect.target.target import Target

CONTROLSET_REGEX = re.compile("CurrentControlSet", flags=re.IGNORECASE)


class RegistryPlugin(Plugin):
    """Provides registry access for Windows targets.

    Acts much the same to how the registry works on a live Windows machine.
    Hives are correctly mapped under e.g. HKLM\\SOFTWARE.

    Internal functions only.
    """

    __namespace__ = "registry"

    SHORTNAMES: Final[dict[str, str]] = {
        "HKLM": "HKEY_LOCAL_MACHINE",
        "HKCC": "HKEY_CURRENT_CONFIG",
        "HKCU": "HKEY_CURRENT_USER",
        "HKCR": "HKEY_CLASSES_ROOT",
        "HKU": "HKEY_USERS",
    }

    MAPPINGS: Final[dict[str, str]] = {
        "SAM": "HKEY_LOCAL_MACHINE\\SAM",
        "SECURITY": "HKEY_LOCAL_MACHINE\\SECURITY",
        "SOFTWARE": "HKEY_LOCAL_MACHINE\\SOFTWARE",
        "SYSTEM": "HKEY_LOCAL_MACHINE\\SYSTEM",
        "COMPONENTS": "HKEY_LOCAL_MACHINE\\COMPONENTS",
        "DEFAULT": "HKEY_USERS\\.DEFAULT",
        "BCD": "HKEY_LOCAL_MACHINE\\BCD00000000",
        "ELAM": "HKEY_LOCAL_MACHINE\\ELAM",
    }

    SYSTEM = (
        "SAM",
        "SECURITY",
        "SOFTWARE",
        "SYSTEM",
        "COMPONENTS",
        "DEFAULT",
        "ELAM",
        "USER.DAT",  # Win 95/98/ME
        "SYSTEM.DAT",  # Win 95/98/ME
        "CLASSES.DAT",  # Win ME
        "REG.DAT",  # Win 3.1
    )

    def __init__(self, target: Target):
        super().__init__(target)

        self._root = VirtualHive()
        self._hive_collections = defaultdict(HiveCollection)
        self._hive_paths = []

        self._hives_to_users: dict[RegfHive, UserDetails] = {}

        self._controlsets = []
        self._currentcontrolset = None
        self._users_loaded = False
        self._init_registry()

        self.key = lru_cache(4096)(self.key)

    def _init_registry(self) -> None:
        dirs = [
            # Windows XP or newer
            ("sysvol/windows/system32/config", False),
            # Windows NT3, NT4, 2k
            ("sysvol/WINNT/system32/config", False),
            # Windows 3.11, 95, 98, ME
            ("sysvol/windows", False),
            # ReactOS (alternative location)
            ("sysvol/reactos", False),
            # RegBack hives are often empty files
            ("sysvol/windows/system32/config/RegBack", True),
        ]
        opened_hives = set()

        for path, log_empty_to_debug in dirs:
            config_dir = self.target.fs.path(path)
            for fname in self.SYSTEM:
                hive_file = config_dir.joinpath(fname)
                if not hive_file.exists():
                    self.target.log.trace("Could not find hive: %s", hive_file)
                    continue

                if hive_file.stat().st_size == 0:
                    # Only log to debug if we have found a valid hive in a previous path
                    log_level = logging.DEBUG if log_empty_to_debug or fname in opened_hives else logging.WARNING
                    self.target.log.log(log_level, "Empty hive: %s", hive_file)
                    continue

                try:
                    hf = RegfHive(hive_file)
                    self._add_hive(fname, hf, hive_file)
                    opened_hives.add(fname)
                except Exception as e:
                    self.target.log.warning("Could not open hive: %s", hive_file)
                    self.target.log.debug("", exc_info=e)

        for drive in ["sysvol", "efi"]:
            bcd = self.target.fs.path(drive).joinpath("boot/BCD")
            if not bcd.exists():
                continue

            if bcd.stat().st_size == 0:
                self.target.log.warning("Empty BCD hive: %s", bcd)
                continue

            try:
                hf = RegfHive(bcd)
                self._add_hive("BCD", hf, bcd)
            except Exception as e:
                self.target.log.warning("Could not open BCD: %s", bcd)
                self.target.log.debug("", exc_info=e)

        for hive, location in self.MAPPINGS.items():
            if hive in self._hive_collections:
                self._map_hive(location, self._hive_collections[hive])

    def _init_users(self) -> None:
        # The initialization of user hives is separated from the initialization
        # of the class on purpose.
        # Loading user hives needs user information (through _os.users()),
        # while getting those users needs a minimal functional registry (the
        # HKLM hive should be available).
        if self._users_loaded:
            return

        for user_details in self.target.user_details.all_with_home():
            user = user_details.user
            ntuser = user_details.home_path.joinpath("ntuser.dat")

            if not ntuser.exists():
                self.target.log.debug("Could not find ntuser.dat: %s", ntuser)
            elif ntuser.stat().st_size == 0:
                self.target.log.warning("Empty NTUSER.DAT hive: %s", ntuser)
            else:
                try:
                    ntuserhive = RegfHive(ntuser)
                    self.add_hive(user.sid, f"HKEY_USERS\\{user.sid}", ntuserhive, ntuser)
                    self._hives_to_users[ntuserhive] = user_details
                except Exception as e:
                    self.target.log.warning("Could not open ntuser.dat: %s", ntuser)
                    self.target.log.debug("", exc_info=e)

            usrclass = user_details.home_path.joinpath("AppData/Local/Microsoft/Windows/usrclass.dat")

            if not usrclass.exists():
                self.target.log.debug("Could not find usrclass.dat: %s", usrclass)
            elif usrclass.stat().st_size == 0:
                self.target.log.warning("Empty UsrClass.DAT hive: %s", usrclass)
            else:
                try:
                    usr_class_hive = RegfHive(usrclass)
                    self.add_hive(f"{user.sid}_Classes", f"HKEY_USERS\\{user.sid}_Classes", usr_class_hive, usrclass)
                    self._map_hive(f"HKEY_USERS\\{user.sid}\\Software\\Classes", usr_class_hive)

                    self._hives_to_users[usr_class_hive] = user_details
                except Exception as e:
                    self.target.log.warning("Could not open usrclass.dat: %s", usrclass)
                    self.target.log.debug("", exc_info=e)

        self._users_loaded = True

    @internal
    def load_user_hives(self) -> None:
        """Load and map the user hives present in the target."""
        self._init_users()

    def _add_hive(self, name: str, hive: RegistryHive, path: TargetPath) -> None:
        """Add a hive to the internal _hive_collections and _hive_paths."""
        self._hive_collections[name.upper()].add(hive)
        self._hive_paths.append((name, hive, path))

    def _map_hive(self, location: str, hive: RegistryHive) -> None:
        """Map a hive to a specific location in the root hive."""
        self._root.map_hive(location, hive)

    @internal
    def add_hive(self, name: str, location: str, hive: RegistryHive, path: TargetPath) -> None:
        """Register and add a hive to a specific location in the root hive."""
        self._add_hive(name, hive, path)
        self._map_hive(location, hive)

    def check_compatible(self) -> None:
        if not len(self._hive_collections):
            raise UnsupportedPluginError("No hive collections found")

    @property
    def controlsets(self) -> list[str]:
        """Return a list of the different ControlSet names."""
        if not self._controlsets:
            for key in self.key("HKLM\\SYSTEM").subkeys():
                if key.name.startswith("ControlSet0"):
                    self._controlsets.append(key.name)
                if key.name == "Select":
                    current_value = key.value("Current").value
                    self._currentcontrolset = f"ControlSet{current_value:03d}"
        return self._controlsets

    @internal
    def root(self) -> KeyCollection:
        """Returns the root of the virtual registry."""
        return self.key()

    @internal
    def key(self, key: str | None = None) -> KeyCollection:
        """Query the virtual registry on the given key.

        Returns a KeyCollection which contains all keys that match
        the query.
        """
        key = (key or "").strip("\\")

        if not key:
            return KeyCollection([self._root.root()])

        if CONTROLSET_REGEX.findall(key):
            if not self._currentcontrolset:
                self.target.log.warning("No known destination for CurrentControlSet link")
                raise RegistryKeyNotFoundError(key)

            key = CONTROLSET_REGEX.sub(self._currentcontrolset, key, 1)

        hive, _, path = key.partition("\\")
        for short_name, name in self.SHORTNAMES.items():
            if hive.upper() == short_name.upper():
                hive = name
                break

        key = f"{hive}\\{path}"

        if hive in ("HKEY_CURRENT_USER", "HKEY_USERS"):
            self._init_users()

        if hive == "HKEY_CURRENT_USER":
            res = KeyCollection()
            for sid in self.key("HKEY_USERS").subkeys():
                if sid.name.endswith("_Classes"):
                    continue
                try:
                    res.add(self.key(f"HKEY_USERS\\{sid.name}\\{path}"))
                except RegistryKeyNotFoundError:
                    pass
            return res
        if hive == "HKEY_CLASSES_ROOT":
            res = self.key(f"HKEY_CURRENT_USER\\Software\\Classes\\{path}")
            res.add(self.key(f"HKEY_LOCAL_MACHINE\\Software\\Classes\\{path}"))
            return res
        res = self._root.key(key)

        if not isinstance(res, KeyCollection):
            res = KeyCollection([res])

        return res

    @internal
    def value(self, key: str, value: str) -> ValueCollection:
        """Convenience method for accessing a specific value."""
        return self.key(key).value(value)

    @internal
    def subkey(self, key: str, subkey: str) -> KeyCollection:
        """Convenience method for accessing a specific subkey."""
        return self.key(key).subkey(subkey)

    @internal
    def keys(self, keys: str | Iterable[str]) -> Iterator[RegistryKey]:
        """Yields all keys that match the given queries.

        Automatically resolves CurrentVersion keys. Also flattens KeyCollections.
        """
        keys = [keys] if isinstance(keys, str) else keys

        for key in self._iter_controlset_keypaths(keys):
            try:
                yield from self.key(key)
            except (RegistryKeyNotFoundError, HiveUnavailableError):  # noqa: PERF203
                pass

    @internal
    def values(self, keys: str | Iterable[str], value: str | Iterable[str]) -> Iterator[RegistryValue]:
        """Yields all values that match the given queries.

        Automatically resolves CurrentVersion keys. Also flattens ValueCollections.
        """

        values = [value] if isinstance(value, str) else value

        for key in self.keys(keys):
            for value in values:
                try:
                    yield key.value(value)
                except RegistryValueNotFoundError:  # noqa: PERF203
                    pass

    def _iter_controlset_keypaths(self, keys: Iterable[str]) -> Iterator[str]:
        """Yield the key transformed for the different control sets."""
        for key in keys:
            if not self.controlsets or not CONTROLSET_REGEX.search(key):
                yield key
                continue

            for controlset in self.controlsets:
                yield CONTROLSET_REGEX.sub(controlset, key)

    @internal
    def iterhives(self) -> Iterator[tuple[str, RegistryHive, TargetPath]]:
        """Returns an iterator for all hives.

        Items are tuples with three members: (name, hive, path)
        """
        return iter(self._hive_paths)

    @internal
    def mappings(self) -> dict[str, str]:
        """Return hive mappings."""
        return self.MAPPINGS

    @internal
    def get_user_details(self, key: RegistryKey | RegistryValue) -> UserDetails | None:
        """Return user details for the user who owns a registry hive that contains the provided key."""
        if not key.hive or not getattr(key.hive, "filepath", None):
            return None

        return self._hives_to_users.get(key.hive)

    @internal
    def get_user(self, key: RegistryKey | RegistryValue) -> WindowsUserRecord | None:
        """Return user record for the user who owns a registry hive that contains the provided key."""
        details = self._hives_to_users.get(key.hive)
        if details:
            return details.user
        return None

    @internal
    def glob_ext(self, pattern: str) -> Iterator[KeyCollection]:
        key_path, pattern = glob_split(pattern)

        try:
            key_collection = self.key(key_path)
        except RegistryKeyNotFoundError:
            return
        else:
            yield from glob_ext(key_collection, pattern)



from io import BytesIO

import pytest

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.os.unix.linux.android._os import AndroidPlugin
from dissect.target.target import Target
from tests._utils import absolute_path


def test_android_os(target_android: Target) -> None:
    target_android.add_plugin(AndroidPlugin)

    assert target_android.os == "android"
    assert target_android.version == "Android 4.4.2 (2013-10-31)"
    assert target_android.hostname == "TMG28071935"


@pytest.mark.parametrize(
    ("build_prop_locations"),
    [
        ([("/build.prop", "build.prop")]),
        ([("/system/build.prop", "build.prop")]),
        ([("/build.prop", "build.prop"), ("/foo/build.prop", "another.prop")]),
    ],
)
def test_android_os_detect_props(target_bare: Target, build_prop_locations: list[tuple[str, str]]) -> None:
    """Test if we detect different build.prop locations correctly."""

    fs = VirtualFilesystem()
    fs.makedirs("/data")
    fs.makedirs("/system")
    fs.makedirs("/vendor")
    fs.makedirs("/product")

    for prop, prop_file in build_prop_locations:
        fs.map_file(prop, absolute_path(f"_data/plugins/os/unix/linux/android/{prop_file}"))

    # prop file that should not be found
    fs.map_file_fh("/foo/bar/too/deep/build.prop", BytesIO(b"ro.not.found='true'"))

    target_bare._os_plugin = AndroidPlugin
    target_bare.filesystems.add(fs)
    target_bare.apply()

    target_bare.add_plugin(AndroidPlugin)

    assert target_bare.os == "android"
    assert sorted(map(str, target_bare._os.build_prop_paths)) == sorted(p for p, _ in build_prop_locations)
    assert target_bare._os.props
    assert target_bare.hostname == "TMG28071935"

    # test if mutual exclusive properties from different build.prop files are added to the dict.
    if "/foo/build.prop" in target_bare._os.build_prop_paths:
        assert target_bare._os.props.get("ro.foo") == "bar"

    # test if glob does not go too deep.
    assert "/foo/bar/too/deep/build.prop" not in target_bare._os.build_prop_paths



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Union

from dissect.database.ese.tools import sru
from dissect.database.exception import Error
from dissect.util.ts import wintimestamp

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

NetworkDataRecord = TargetRecordDescriptor(
    "filesystem/windows/sru/network_data",
    [
        ("datetime", "ts"),
        ("path", "app"),
        ("string", "user"),
        ("varint", "interface_luid"),
        ("varint", "l2_profile_id"),
        ("varint", "l2_profile_flags"),
        ("varint", "bytes_sent"),
        ("varint", "bytes_recvd"),
    ],
)

NetworkConnectivityRecord = TargetRecordDescriptor(
    "filesystem/windows/sru/network_connectivity",
    [
        ("datetime", "ts"),
        ("path", "app"),
        ("string", "user"),
        ("varint", "interface_luid"),
        ("varint", "l2_profile_id"),
        ("varint", "connected_time"),
        ("datetime", "connect_start_time"),
        ("varint", "l2_profile_flags"),
    ],
)

EnergyEstimatorRecord = TargetRecordDescriptor(
    "filesystem/windows/sru/energy_estimator",
    [
        ("datetime", "ts"),
        ("path", "app"),
        ("string", "user"),
        ("bytes", "binary_data"),
    ],
)

EnergyUsageRecord = TargetRecordDescriptor(
    "filesystem/windows/sru/energy_usage",
    [
        ("datetime", "ts"),
        ("path", "app"),
        ("string", "user"),
        ("varint", "event_timestamp"),
        ("varint", "state_transition"),
        ("varint", "designed_capacity"),
        ("varint", "full_charged_capacity"),
        ("varint", "charge_level"),
        ("varint", "cycle_count"),
        ("varint", "configuration_hash"),
        ("varint", "battery_count"),
        ("varint", "battery_charge_limited"),
    ],
)

EnergyUsageLTRecord = TargetRecordDescriptor(
    "filesystem/windows/sru/energy_usage_lt",
    [
        ("datetime", "ts"),
        ("path", "app"),
        ("string", "user"),
        ("varint", "active_ac_time"),
        ("varint", "cs_ac_time"),
        ("varint", "active_dc_time"),
        ("varint", "cs_dc_time"),
        ("varint", "active_discharge_time"),
        ("varint", "cs_discharge_time"),
        ("varint", "active_energy"),
        ("varint", "cs_energy"),
        ("varint", "designed_capacity"),
        ("varint", "full_charged_capacity"),
        ("varint", "cycle_count"),
        ("varint", "configuration_hash"),
    ],
)

ApplicationRecord = TargetRecordDescriptor(
    "filesystem/windows/sru/application",
    [
        ("datetime", "ts"),
        ("path", "app"),
        ("string", "user"),
        ("varint", "foreground_cycle_time"),
        ("varint", "background_cycle_time"),
        ("varint", "face_time"),
        ("varint", "foreground_context_switches"),
        ("varint", "background_context_switches"),
        ("varint", "foreground_bytes_read"),
        ("varint", "foreground_bytes_written"),
        ("varint", "foreground_num_read_operations"),
        ("varint", "foreground_num_write_operations"),
        ("varint", "foreground_number_of_flushes"),
        ("varint", "background_bytes_read"),
        ("varint", "background_bytes_written"),
        ("varint", "background_num_read_operations"),
        ("varint", "background_num_write_operations"),
        ("varint", "background_number_of_flushes"),
    ],
)

PushNotificationRecord = TargetRecordDescriptor(
    "filesystem/windows/sru/push_notification",
    [
        ("datetime", "ts"),
        ("path", "app"),
        ("string", "user"),
        ("varint", "notification_type"),
        ("varint", "payload_size"),
        ("varint", "network_type"),
    ],
)

ApplicationTimelineRecord = TargetRecordDescriptor(
    "filesystem/windows/sru/application_timeline",
    [
        ("datetime", "ts"),
        ("path", "app"),
        ("string", "user"),
        ("varint", "flags"),
        ("datetime", "end_time"),
        ("varint", "duration_ms"),
        ("varint", "span_ms"),
        ("varint", "timeline_end"),
        ("varint", "in_focus_timeline"),
        ("varint", "user_input_timeline"),
        ("varint", "comp_rendered_timeline"),
        ("varint", "comp_dirtied_timeline"),
        ("varint", "comp_propagated_timeline"),
        ("varint", "audio_in_timeline"),
        ("varint", "audio_out_timeline"),
        ("varint", "cpu_timeline"),
        ("varint", "disk_timeline"),
        ("varint", "network_timeline"),
        ("varint", "mbb_timeline"),
        ("varint", "in_focus_s"),
        ("varint", "psm_foreground_s"),
        ("varint", "user_input_s"),
        ("varint", "comp_rendered_s"),
        ("varint", "comp_dirtied_s"),
        ("varint", "comp_propagated_s"),
        ("varint", "audio_in_s"),
        ("varint", "audio_out_s"),
        ("varint", "cycles"),
        ("varint", "cycles_breakdown"),
        ("varint", "cycles_attr"),
        ("varint", "cycles_attr_breakdown"),
        ("varint", "cycles_wob"),
        ("varint", "cycles_wob_breakdown"),
        ("varint", "disk_raw"),
        ("varint", "network_tail_raw"),
        ("varint", "network_bytes_raw"),
        ("varint", "mbb_tail_raw"),
        ("varint", "mbb_bytes_raw"),
        ("varint", "display_required_s"),
        ("varint", "display_required_timeline"),
        ("varint", "keyboard_input_timeline"),
        ("varint", "keyboard_input_s"),
        ("varint", "mouse_input_s"),
    ],
)

VfuRecord = TargetRecordDescriptor(
    "filesystem/windows/sru/vfu",
    [
        ("datetime", "ts"),
        ("path", "app"),
        ("string", "user"),
        ("varint", "flags"),
        ("datetime", "start_time"),
        ("datetime", "end_time"),
        ("bytes", "usage"),
    ],
)

SdpVolumeProviderRecord = TargetRecordDescriptor(
    "filesystem/windows/sru/sdp_volume_provider",
    [
        ("datetime", "ts"),
        ("path", "app"),
        ("string", "user"),
        ("varint", "total"),
        ("varint", "used"),
    ],
)

SdpPhysicalDiskProviderRecord = TargetRecordDescriptor(
    "filesystem/windows/sru/sdp_physical_disk_provider",
    [
        ("datetime", "ts"),
        ("path", "app"),
        ("string", "user"),
        ("varint", "size_in_bytes"),
    ],
)

SdpCpuProviderRecord = TargetRecordDescriptor(
    "filesystem/windows/sru/sdp_cpu_provider",
    [
        ("datetime", "ts"),
        ("path", "app"),
        ("string", "user"),
        ("varint", "processor_time"),
    ],
)

SdpNetworkProviderRecord = TargetRecordDescriptor(
    "filesystem/windows/sru/sdp_network_provider",
    [
        ("datetime", "ts"),
        ("path", "app"),
        ("string", "user"),
        ("varint", "bytes_inbound"),
        ("varint", "bytes_outbound"),
        ("varint", "bytes_total"),
    ],
)

SRURecord = Union[  # noqa: UP007
    NetworkDataRecord,
    NetworkConnectivityRecord,
    EnergyEstimatorRecord,
    EnergyUsageRecord,
    EnergyUsageLTRecord,
    ApplicationRecord,
    PushNotificationRecord,
    ApplicationTimelineRecord,
    VfuRecord,
    SdpVolumeProviderRecord,
    SdpPhysicalDiskProviderRecord,
    SdpCpuProviderRecord,
    SdpNetworkProviderRecord,
]
FIELD_MAPPINGS = {
    "ActiveAcTime": "active_ac_time",
    "ActiveDcTime": "active_dc_time",
    "ActiveDischargeTime": "active_discharge_time",
    "ActiveEnergy": "active_energy",
    "AppId": "app",
    "AudioInS": "audio_in_s",
    "AudioInTimeline": "audio_in_timeline",
    "AudioOutS": "audio_out_s",
    "AudioOutTimeline": "audio_out_timeline",
    "BackgroundBytesRead": "background_bytes_read",
    "BackgroundBytesWritten": "background_bytes_written",
    "BackgroundContextSwitches": "background_context_switches",
    "BackgroundCycleTime": "background_cycle_time",
    "BackgroundNumReadOperations": "background_num_read_operations",
    "BackgroundNumWriteOperations": "background_num_write_operations",
    "BackgroundNumberOfFlushes": "background_number_of_flushes",
    "BatteryChargeLimited": "battery_charge_limited",
    "BatteryCount": "battery_count",
    "BinaryData": "binary_data",
    "BytesInBound": "bytes_inbound",
    "BytesOutBound": "bytes_outbound",
    "BytesRecvd": "bytes_recvd",
    "BytesSent": "bytes_sent",
    "BytesTotal": "bytes_total",
    "ChargeLevel": "charge_level",
    "CompDirtiedS": "comp_dirtied_s",
    "CompDirtiedTimeline": "comp_dirtied_timeline",
    "CompPropagatedS": "comp_propagated_s",
    "CompPropagatedTimeline": "comp_propagated_timeline",
    "CompRenderedS": "comp_rendered_s",
    "CompRenderedTimeline": "comp_rendered_timeline",
    "ConfigurationHash": "configuration_hash",
    "ConnectStartTime": "connect_start_time",
    "ConnectedTime": "connected_time",
    "CpuTimeline": "cpu_timeline",
    "CsAcTime": "cs_ac_time",
    "CsDcTime": "cs_dc_time",
    "CsDischargeTime": "cs_discharge_time",
    "CsEnergy": "cs_energy",
    "CycleCount": "cycle_count",
    "Cycles": "cycles",
    "CyclesAttr": "cycles_attr",
    "CyclesAttrBreakdown": "cycles_attr_breakdown",
    "CyclesBreakdown": "cycles_breakdown",
    "CyclesWOB": "cycles_wob",
    "CyclesWOBBreakdown": "cycles_wob_breakdown",
    "DesignedCapacity": "designed_capacity",
    "DiskRaw": "disk_raw",
    "DiskTimeline": "disk_timeline",
    "DisplayRequiredS": "display_required_s",
    "DisplayRequiredTimeline": "display_required_timeline",
    "DurationMS": "duration_ms",
    "EndTime": "end_time",
    "EventTimestamp": "event_timestamp",
    "FaceTime": "face_time",
    "Flags": "flags",
    "ForegroundBytesRead": "foreground_bytes_read",
    "ForegroundBytesWritten": "foreground_bytes_written",
    "ForegroundContextSwitches": "foreground_context_switches",
    "ForegroundCycleTime": "foreground_cycle_time",
    "ForegroundNumReadOperations": "foreground_num_read_operations",
    "ForegroundNumWriteOperations": "foreground_num_write_operations",
    "ForegroundNumberOfFlushes": "foreground_number_of_flushes",
    "FullChargedCapacity": "full_charged_capacity",
    "InFocusS": "in_focus_s",
    "InFocusTimeline": "in_focus_timeline",
    "InterfaceLuid": "interface_luid",
    "KeyboardInputS": "keyboard_input_s",
    "KeyboardInputTimeline": "keyboard_input_timeline",
    "L2ProfileFlags": "l2_profile_flags",
    "L2ProfileId": "l2_profile_id",
    "MBBBytesRaw": "mbb_bytes_raw",
    "MBBTailRaw": "mbb_tail_raw",
    "MBBTimeline": "mbb_timeline",
    "MouseInputS": "mouse_input_s",
    "NetworkBytesRaw": "network_bytes_raw",
    "NetworkTailRaw": "network_tail_raw",
    "NetworkTimeline": "network_timeline",
    "NetworkType": "network_type",
    "NotificationType": "notification_type",
    "PSMForegroundS": "psm_foreground_s",
    "PayloadSize": "payload_size",
    "ProcessorTime": "processor_time",
    "SizeInBytes": "size_in_bytes",
    "SpanMS": "span_ms",
    "StartTime": "start_time",
    "StateTransition": "state_transition",
    "TimeStamp": "ts",
    "TimelineEnd": "timeline_end",
    "Total": "total",
    "Usage": "usage",
    "Used": "used",
    "UserId": "user",
    "UserInputS": "user_input_s",
    "UserInputTimeline": "user_input_timeline",
}


def transform_app_id(value: bytes | str | None) -> str | None:
    if value is not None:
        value = value.decode() if isinstance(value, bytes) else str(value)
    return value


def transform_timestamp(value: int | datetime) -> datetime:
    if isinstance(value, datetime):
        return value
    return wintimestamp(value)


TRANSFORMS = {
    "AppId": transform_app_id,
    "StartTime": transform_timestamp,
    "EndTime": transform_timestamp,
}


class SRUPlugin(Plugin):
    """Return all available SRUM data stored in the SRUDB.dat.

    The System Resource Usage Monitor (SRUM) stores its information in a SRUDB.dat file. As the names suggests, it
    contains data about resource usage, such as network and memory usage by applications.

    References:
        - https://docs.microsoft.com/en-us/sql/relational-databases/performance-monitor/monitor-resource-usage-system-monitor?view=sql-server-ver15
        - https://blog.1234n6.com/2019/01/
    """

    __namespace__ = "sru"

    def __init__(self, target: Target):
        super().__init__(target)
        self._sru = None

        srupath = target.resolve("%windir%/System32/sru/SRUDB.dat")
        if srupath.exists():
            try:
                self._sru = sru.SRU(srupath.open())
            except Error as e:
                self.target.log.warning("Error opening SRU database")
                self.target.log.debug("", exc_info=e)

    def check_compatible(self) -> None:
        if not self._sru:
            raise UnsupportedPluginError("No SRUDB found")

    def read_records(self, table_name: str, record_type: SRURecord) -> Iterator[SRURecord]:
        table = self._sru.get_table(table_name=table_name)
        if not table:
            self.target.log.warning("Table not found: %s", table_name)
            return iter(())

        columns = [c.name for c in table.columns]
        if columns[:4] != ["AutoIncId", "TimeStamp", "AppId", "UserId"]:
            raise ValueError(f"Unexpected table layout in SRU iteration: {table} ({columns[:4]})")
        columns = columns[1:]

        for entry in self._sru.get_table_entries(table=table):
            values = (entry[name] for name in columns)
            column_values = zip(columns, values, strict=False)

            record_values = {}
            for column, value in column_values:
                new_value = value
                if new_value and (transform := TRANSFORMS.get(column)):
                    if isinstance((transformed_value := transform(new_value)), str):
                        new_value = self.target.fs.path(transformed_value)
                    else:
                        new_value = transformed_value
                new_column = FIELD_MAPPINGS.get(column, column)
                record_values[new_column] = new_value

            yield record_type(
                _target=self.target,
                **record_values,
            )

    @export(record=NetworkDataRecord)
    def network_data(self) -> Iterator[NetworkDataRecord]:
        """Return the contents of Windows Network Data Usage Monitor table from the SRUDB.dat file.

        Gives insight into the network usage of the system.
        """
        yield from self.read_records("network_data", NetworkDataRecord)

    @export(record=NetworkConnectivityRecord)
    def network_connectivity(self) -> Iterator[NetworkConnectivityRecord]:
        """Return the contents of Windows Network Connectivity Usage Monitor table from the SRUDB.dat file.

        Gives insight into the network connectivity usage of the system.
        """
        yield from self.read_records("network_connectivity", NetworkConnectivityRecord)

    @export(record=EnergyEstimatorRecord)
    def energy_estimator(self) -> Iterator[EnergyEstimatorRecord]:
        """Return the contents of Energy Estimator table from the SRUDB.dat file."""
        yield from self.read_records("energy_estimator", EnergyEstimatorRecord)

    @export(record=EnergyUsageRecord)
    def energy_usage(self) -> Iterator[EnergyUsageRecord]:
        """Return the contents of Energy Usage Provider table from the SRUDB.dat file.

        Gives insight into the energy usage of the system.
        """
        yield from self.read_records("energy_usage", EnergyUsageRecord)

    @export(record=EnergyUsageLTRecord)
    def energy_usage_lt(self) -> Iterator[EnergyUsageLTRecord]:
        """Return the contents of Energy Usage Provider Long Term table from the SRUDB.dat file.

        Gives insight into the energy usage of the system looking over the long term.
        """
        yield from self.read_records("energy_usage_lt", EnergyUsageLTRecord)

    @export(record=ApplicationRecord)
    def application(self) -> Iterator[ApplicationRecord]:
        """Return the contents of Application Resource Usage table from the SRUDB.dat file.

        Gives insights into the resource usage of applications on the system.
        """
        yield from self.read_records("application", ApplicationRecord)

    @export(record=PushNotificationRecord)
    def push_notification(self) -> Iterator[PushNotificationRecord]:
        """Return the contents of Windows Push Notification Data table from the SRUDB.dat file.

        Gives insight into the notification usage of the system.
        """
        yield from self.read_records("push_notifications", PushNotificationRecord)

    @export(record=ApplicationTimelineRecord)
    def application_timeline(self) -> Iterator[ApplicationTimelineRecord]:
        """Return the contents of App Timeline Provider table from the SRUDB.dat file."""
        yield from self.read_records("application_timeline", ApplicationTimelineRecord)

    @export(record=VfuRecord)
    def vfu(self) -> Iterator[VfuRecord]:
        """Return the contents of vfuprov table from the SRUDB.dat file."""
        yield from self.read_records("vfu", VfuRecord)

    @export(record=SdpVolumeProviderRecord)
    def sdp_volume_provider(self) -> Iterator[SdpVolumeProviderRecord]:
        """Return the contents of SDP Volume Provider table from the SRUDB.dat file."""
        yield from self.read_records("sdp_volume_provider", SdpVolumeProviderRecord)

    @export(record=SdpPhysicalDiskProviderRecord)
    def sdp_physical_disk_provider(self) -> Iterator[SdpPhysicalDiskProviderRecord]:
        """Return the contents of SDP Physical Disk Provider table from the SRUDB.dat file."""
        yield from self.read_records("sdp_physical_disk_provider", SdpPhysicalDiskProviderRecord)

    @export(record=SdpCpuProviderRecord)
    def sdp_cpu_provider(self) -> Iterator[SdpCpuProviderRecord]:
        """Return the contents of SDP CPU Provider table from the SRUDB.dat file."""
        yield from self.read_records("sdp_cpu_provider", SdpCpuProviderRecord)

    @export(record=SdpNetworkProviderRecord)
    def sdp_network_provider(self) -> Iterator[SdpNetworkProviderRecord]:
        """Return the contents of SDP Network Provider table from the SRUDB.dat file."""
        yield from self.read_records("sdp_network_provider", SdpNetworkProviderRecord)



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.evidence.ad1.ad1 import find_files

from dissect.target.filesystems.ad1 import AD1Filesystem
from dissect.target.loader import Loader

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target


class AD1Loader(Loader):
    """Access Data ``.ad`` loader."""

    @staticmethod
    def detect(path: Path) -> bool:
        return path.suffix.lower() == ".ad1"

    def map(self, target: Target) -> None:
        """Map the detected segment files as an :class:`AD1Filesystem` to the target.

        Currently does not detect NTFS / case-insensitive filesystems or custom content
        images with multiple sources.
        """
        try:
            fs = AD1Filesystem(find_files(self.path))
            target.filesystems.add(fs)

        except ValueError as e:
            target.log.error("Unable to map AD1: %s", e)  # noqa: TRY400



from __future__ import annotations

import shutil
import sys

import dissect.extfs.exceptions
import dissect.ntfs.exceptions
import dissect.xfs.exceptions

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.plugin import Plugin, arg, export


class ICatPlugin(Plugin):
    """Plugin to output the contents of a file based on its MFT segment or inode number."""

    FS_SUPPORTED = ("ntfs", "xfs", "ext", "virtual")

    def check_compatible(self) -> None:
        filesystems = self.target.filesystems
        if not any(fs.__type__ in self.FS_SUPPORTED for fs in filesystems):
            raise UnsupportedPluginError("No supported filesystems found")

    @arg("-i", "--inode", "--segment", dest="inum", type=int, required=True, help="MFT segment or inode number")
    @arg(
        "--fs",
        type=int,
        help="optional filesystem index, zero indexed. Defaults to the 'sysvol' or '/' filesystem otherwise",
    )
    @arg("--ads", default="", help="Alternate Data Stream name")
    @export(output="none")
    def icat(self, inum: int, fs: int | None, ads: str) -> None:
        """Output the contents of a file based on its MFT segment or inode number. Supports Alternate Data Streams

        Example:
            .. code-block::

                # outputs contents of segment defaults to 'sysvol'
                target-query <TARGET> -f icat --segment 96997

                # outputs contents of inode defaults to '/'
                target-query <TARGET> -f icat --inode 50947

                # outputs contents of segment's ADS
                target-query <TARGET> -f icat --segment 96997 --ads Zone.Identifier

                # outputs contents of segment in filesystem 3 of target
                target-query <TARGET> -f icat --fs 3 --segment 96997

                # outputs contents of inode in filesystem 2 of target
                target-query <TARGET> -f icat --fs 2 --inode 50947
        """

        open_as = None
        try:
            if fs is not None:
                try:
                    filesystem = self.target.filesystems[fs]
                except IndexError:
                    self.target.log.exception("%s does not have a filesystem with index number: %s", self.target, fs)
                    return
            else:
                if "sysvol" in self.target.fs.mounts:
                    filesystem = self.target.fs.mounts["sysvol"]
                    # In some cases the fstype of sysvol can be virtual. For this
                    # case we set the open_as sentinel so the code
                    # opening the filesystem handle know how to open it.
                    open_as = "ntfs"
                elif "/" in self.target.fs.mounts:
                    filesystem = self.target.fs.mounts["/"]
                else:
                    self.target.log.exception(
                        '%s does not contain mountpoints "sysvol" or "/" '
                        'specify your own filesystem using the "--fs" option',
                        self.target,
                    )
                    return

            if filesystem.__type__ == "ntfs" or open_as == "ntfs":
                fh = filesystem.ntfs.mft(inum).open(ads)
            elif filesystem.__type__ == "ext":
                fh = filesystem.extfs.get_inode(inum).open()
            elif filesystem.__type__ == "xfs":
                fh = filesystem.xfs.get_inode(inum).open()
            else:
                self.target.log.exception('Unsupported FS type "%s"', filesystem.__type__)
                return

            shutil.copyfileobj(fh, sys.stdout.buffer)

        except (
            dissect.ntfs.exceptions.Error,
            dissect.extfs.exceptions.Error,
            dissect.xfs.exceptions.Error,
        ):
            # For example, ntfs.exceptions.VolumeNotAvailableError that gets thrown
            # when you try to icat a file from an acquire which is non-resident
            self.target.log.exception(
                "%s failed to get contents of file with segment / inode number %s from filesystem %s (ADS: %s)",
                self.target,
                inum,
                filesystem,
                ads,
            )
